"""Async payments client wrapping Stripe."""

from __future__ import annotations

import inspect
from typing import Any, cast

from etils import epy

from derp.config import PaymentsConfig
from derp.payments.exceptions import (
    PaymentsNotConnectedError,
    PaymentsProviderError,
    WebhookSignatureError,
)
from derp.payments.models import (
    Account,
    AccountLink,
    AccountLinkType,
    AccountType,
    Balance,
    CancellationReason,
    CaptureMethod,
    CheckoutSession,
    CheckoutSessionMode,
    Customer,
    PaymentIntent,
    Payout,
    PayoutMethod,
    Refund,
    RefundReason,
    StripeListResult,
    Transfer,
    WebhookEvent,
)

with epy.lazy_imports():
    import stripe


class PaymentsClient:
    """Async wrapper around Stripe payments APIs."""

    def __init__(self, config: PaymentsConfig):
        self._config = config
        self._stripe_client: stripe.StripeClient | None = None
        self._http_client: stripe.HTTPXClient | None = None

    async def connect(self) -> None:
        """Initialize Stripe client state."""
        if self._stripe_client is not None:
            return

        self._http_client = stripe.HTTPXClient(timeout=self._config.timeout_seconds)
        self._stripe_client = stripe.StripeClient(
            self._config.api_key,
            max_network_retries=self._config.max_network_retries,
            http_client=self._http_client,
        )

    async def disconnect(self) -> None:
        """Clear Stripe client state."""
        if self._http_client is not None:
            close_async = getattr(self._http_client, "close_async", None)
            if callable(close_async):
                maybe_awaitable = close_async()
                if inspect.isawaitable(maybe_awaitable):
                    await maybe_awaitable

        self._stripe_client = None
        self._http_client = None

    async def _provider_call(self, method: Any, *args: Any, **kwargs: Any) -> Any:
        try:
            return await method(*args, **kwargs)
        except Exception as exc:  # pragma: no cover - behavior validated via tests
            raise PaymentsProviderError(
                str(exc) or "Payments provider request failed",
                code=str(getattr(exc, "code", "payments_provider_error")),
            ) from exc

    @staticmethod
    def _to_raw(payload: Any) -> dict[str, Any]:
        if payload is None:
            return {}
        if isinstance(payload, dict):
            return dict(payload)
        values = getattr(payload, "__dict__", None)
        if isinstance(values, dict):
            return dict(values)
        return {}

    @staticmethod
    def _to_metadata(value: Any) -> dict[str, str]:
        if not isinstance(value, dict):
            return {}
        metadata: dict[str, str] = {}
        for key, item in value.items():
            metadata[str(key)] = str(item)
        return metadata

    @staticmethod
    def _customer_id(value: Any) -> str | None:
        if isinstance(value, str):
            return value
        if isinstance(value, dict) and isinstance(value.get("id"), str):
            return cast(str, value["id"])
        return None

    @staticmethod
    def _normalize_customer(raw: dict[str, Any]) -> Customer:
        return Customer(
            id=str(raw.get("id", "")),
            email=cast(str | None, raw.get("email")),
            name=cast(str | None, raw.get("name")),
            phone=cast(str | None, raw.get("phone")),
            metadata=PaymentsClient._to_metadata(raw.get("metadata")),
            created=cast(int | None, raw.get("created")),
            raw=raw,
        )

    @staticmethod
    def _normalize_checkout_session(raw: dict[str, Any]) -> CheckoutSession:
        mode: CheckoutSessionMode | None = None
        if raw.get("mode") in ("payment", "subscription"):
            mode = cast(CheckoutSessionMode, raw["mode"])

        return CheckoutSession(
            id=str(raw.get("id", "")),
            url=cast(str | None, raw.get("url")),
            mode=mode,
            customer_id=PaymentsClient._customer_id(raw.get("customer")),
            customer_email=cast(str | None, raw.get("customer_email")),
            payment_status=cast(str | None, raw.get("payment_status")),
            status=cast(str | None, raw.get("status")),
            expires_at=cast(int | None, raw.get("expires_at")),
            raw=raw,
        )

    @staticmethod
    def _normalize_event(raw: dict[str, Any]) -> WebhookEvent:
        data_object: dict[str, Any] | None = None
        data = raw.get("data")
        if isinstance(data, dict):
            obj = data.get("object")
            if isinstance(obj, dict):
                data_object = obj
            else:
                maybe_object = PaymentsClient._to_raw(obj)
                if maybe_object:
                    data_object = maybe_object

        return WebhookEvent(
            id=str(raw.get("id", "")),
            type=str(raw.get("type", "")),
            created=int(raw.get("created", 0)),
            livemode=bool(raw.get("livemode", False)),
            data_object=data_object,
            raw=raw,
        )

    @staticmethod
    def _connect_options(stripe_account: str | None) -> dict[str, Any] | None:
        if stripe_account is None:
            return None
        return {"stripe_account": stripe_account}

    @staticmethod
    def _normalize_account(raw: dict[str, Any]) -> Account:
        return Account(
            id=str(raw.get("id", "")),
            type=cast(str | None, raw.get("type")),
            email=cast(str | None, raw.get("email")),
            country=cast(str | None, raw.get("country")),
            charges_enabled=bool(raw.get("charges_enabled", False)),
            payouts_enabled=bool(raw.get("payouts_enabled", False)),
            details_submitted=bool(raw.get("details_submitted", False)),
            business_type=cast(str | None, raw.get("business_type")),
            metadata=PaymentsClient._to_metadata(raw.get("metadata")),
            created=cast(int | None, raw.get("created")),
            raw=raw,
        )

    @staticmethod
    def _normalize_account_link(raw: dict[str, Any]) -> AccountLink:
        return AccountLink(
            url=str(raw.get("url", "")),
            created=cast(int | None, raw.get("created")),
            expires_at=cast(int | None, raw.get("expires_at")),
            raw=raw,
        )

    @staticmethod
    def _normalize_transfer(raw: dict[str, Any]) -> Transfer:
        return Transfer(
            id=str(raw.get("id", "")),
            amount=int(raw.get("amount", 0)),
            currency=str(raw.get("currency", "")),
            destination=cast(str | None, raw.get("destination")),
            description=cast(str | None, raw.get("description")),
            transfer_group=cast(str | None, raw.get("transfer_group")),
            metadata=PaymentsClient._to_metadata(raw.get("metadata")),
            created=cast(int | None, raw.get("created")),
            raw=raw,
        )

    @staticmethod
    def _normalize_payment_intent(raw: dict[str, Any]) -> PaymentIntent:
        return PaymentIntent(
            id=str(raw.get("id", "")),
            amount=int(raw.get("amount", 0)),
            currency=str(raw.get("currency", "")),
            status=cast(str | None, raw.get("status")),
            customer_id=PaymentsClient._customer_id(raw.get("customer")),
            description=cast(str | None, raw.get("description")),
            capture_method=cast(str | None, raw.get("capture_method")),
            cancellation_reason=cast(str | None, raw.get("cancellation_reason")),
            payment_method=cast(str | None, raw.get("payment_method")),
            metadata=PaymentsClient._to_metadata(raw.get("metadata")),
            created=cast(int | None, raw.get("created")),
            raw=raw,
        )

    @staticmethod
    def _normalize_refund(raw: dict[str, Any]) -> Refund:
        return Refund(
            id=str(raw.get("id", "")),
            amount=int(raw.get("amount", 0)),
            currency=str(raw.get("currency", "")),
            status=cast(str | None, raw.get("status")),
            payment_intent_id=cast(str | None, raw.get("payment_intent")),
            charge_id=cast(str | None, raw.get("charge")),
            reason=cast(str | None, raw.get("reason")),
            metadata=PaymentsClient._to_metadata(raw.get("metadata")),
            created=cast(int | None, raw.get("created")),
            raw=raw,
        )

    @staticmethod
    def _normalize_payout(raw: dict[str, Any]) -> Payout:
        return Payout(
            id=str(raw.get("id", "")),
            amount=int(raw.get("amount", 0)),
            currency=str(raw.get("currency", "")),
            status=cast(str | None, raw.get("status")),
            method=cast(str | None, raw.get("method")),
            description=cast(str | None, raw.get("description")),
            destination=cast(str | None, raw.get("destination")),
            metadata=PaymentsClient._to_metadata(raw.get("metadata")),
            arrival_date=cast(int | None, raw.get("arrival_date")),
            created=cast(int | None, raw.get("created")),
            raw=raw,
        )

    @staticmethod
    def _normalize_balance(raw: dict[str, Any]) -> Balance:
        available = raw.get("available")
        pending = raw.get("pending")
        connect_reserved = raw.get("connect_reserved")
        return Balance(
            available=list(available) if isinstance(available, list) else [],
            pending=list(pending) if isinstance(pending, list) else [],
            connect_reserved=(
                list(connect_reserved) if isinstance(connect_reserved, list) else None
            ),
            raw=raw,
        )

    @staticmethod
    def _normalize_line_items(
        line_items: list[dict[str, str | int]],
    ) -> list[dict[str, str | int]]:
        if not line_items:
            raise ValueError("line_items must contain at least one item.")

        normalized: list[dict[str, str | int]] = []
        for index, item in enumerate(line_items):
            if "price_id" not in item:
                raise ValueError(
                    f"line_items[{index}] must include a 'price_id' field."
                )
            if "quantity" not in item:
                raise ValueError(
                    f"line_items[{index}] must include a 'quantity' field."
                )

            price_id = item["price_id"]
            quantity = item["quantity"]

            if not isinstance(price_id, str) or not price_id.strip():
                raise ValueError(
                    f"line_items[{index}].price_id must be a non-empty str."
                )
            if (
                isinstance(quantity, bool)
                or not isinstance(quantity, int)
                or quantity <= 0
            ):
                raise ValueError(
                    f"line_items[{index}].quantity must be a positive int."
                )

            normalized.append({"price": price_id, "quantity": quantity})

        return normalized

    async def create_customer(
        self,
        *,
        email: str | None = None,
        name: str | None = None,
        phone: str | None = None,
        metadata: dict[str, str] | None = None,
    ) -> Customer:
        """Create a Stripe customer."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()

        params: dict[str, Any] = {}
        if email is not None:
            params["email"] = email
        if name is not None:
            params["name"] = name
        if phone is not None:
            params["phone"] = phone
        if metadata is not None:
            params["metadata"] = metadata

        response = await self._provider_call(
            self._stripe_client.v1.customers.create_async, params
        )
        return self._normalize_customer(self._to_raw(response))

    async def retrieve_customer(
        self, customer_id: str, *, expand: list[str] | None = None
    ) -> Customer:
        """Retrieve a Stripe customer."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()

        params: dict[str, Any] = {}
        if expand is not None:
            params["expand"] = expand

        response = await self._provider_call(
            self._stripe_client.v1.customers.retrieve_async,
            customer_id,
            params,
        )
        return self._normalize_customer(self._to_raw(response))

    async def update_customer(
        self,
        customer_id: str,
        *,
        email: str | None = None,
        name: str | None = None,
        phone: str | None = None,
        metadata: dict[str, str] | None = None,
    ) -> Customer:
        """Update a Stripe customer."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()

        params: dict[str, Any] = {}
        if email is not None:
            params["email"] = email
        if name is not None:
            params["name"] = name
        if phone is not None:
            params["phone"] = phone
        if metadata is not None:
            params["metadata"] = metadata

        response = await self._provider_call(
            self._stripe_client.v1.customers.update_async,
            customer_id,
            params,
        )
        return self._normalize_customer(self._to_raw(response))

    async def create_checkout_session(
        self,
        *,
        mode: CheckoutSessionMode | str,
        success_url: str,
        cancel_url: str,
        line_items: list[dict[str, str | int]],
        customer_id: str | None = None,
        customer_email: str | None = None,
        client_reference_id: str | None = None,
        metadata: dict[str, str] | None = None,
        allow_promotion_codes: bool | None = None,
        idempotency_key: str | None = None,
    ) -> CheckoutSession:
        """Create a Stripe checkout session."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        if mode not in CheckoutSessionMode:
            raise ValueError(
                f"Invalid mode: {mode}. Must be one of {list(CheckoutSessionMode)}."
            )
        if customer_id is not None and customer_email is not None:
            raise ValueError(
                "Only one of `customer_id` and `customer_email` can be provided,"
                " but both were provided."
            )

        normalized_line_items = self._normalize_line_items(line_items)

        params: dict[str, Any] = {
            "mode": mode,
            "success_url": success_url,
            "cancel_url": cancel_url,
            "line_items": normalized_line_items,
        }

        if customer_id is not None:
            params["customer"] = customer_id
        if customer_email is not None:
            params["customer_email"] = customer_email
        if client_reference_id is not None:
            params["client_reference_id"] = client_reference_id
        if metadata is not None:
            params["metadata"] = metadata
        if allow_promotion_codes is not None:
            params["allow_promotion_codes"] = allow_promotion_codes

        options: dict[str, Any] | None = None
        if idempotency_key is not None:
            if not isinstance(idempotency_key, str) or not idempotency_key.strip():
                raise ValueError(
                    "idempotency_key must be a non-empty str when provided."
                )
            options = {"idempotency_key": idempotency_key}

        args: list[Any] = [params]
        if options is not None:
            args.append(options)

        response = await self._provider_call(
            self._stripe_client.v1.checkout.sessions.create_async, *args
        )
        return self._normalize_checkout_session(self._to_raw(response))

    async def retrieve_checkout_session(
        self, session_id: str, *, expand: list[str] | None = None
    ) -> CheckoutSession:
        """Retrieve a Stripe checkout session."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()

        params: dict[str, Any] = {}
        if expand is not None:
            params["expand"] = expand

        response = await self._provider_call(
            self._stripe_client.v1.checkout.sessions.retrieve_async,
            session_id,
            params,
        )
        return self._normalize_checkout_session(self._to_raw(response))

    async def expire_checkout_session(self, session_id: str) -> CheckoutSession:
        """Expire a Stripe checkout session."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()

        response = await self._provider_call(
            self._stripe_client.v1.checkout.sessions.expire_async,
            session_id,
        )
        return self._normalize_checkout_session(self._to_raw(response))

    async def _list_resource(
        self,
        list_method: Any,
        *,
        limit: int = 25,
        starting_after: str | None = None,
        expand: list[str] | None = None,
        extra_params: dict[str, Any] | None = None,
        options: dict[str, Any] | None = None,
    ) -> StripeListResult:
        """List Stripe resources with cursor pagination."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()

        limit = min(max(limit, 1), 100)
        params: dict[str, Any] = {"limit": limit}
        if starting_after is not None:
            params["starting_after"] = starting_after
        if expand is not None:
            params["expand"] = expand
        if extra_params is not None:
            params.update(extra_params)

        if options is not None:
            result = await self._provider_call(list_method, params, options)
        else:
            result = await self._provider_call(list_method, params)
        data = [self._to_raw(item) for item in result.data]
        return StripeListResult(data=data, has_more=result.has_more)

    async def list_customers(
        self,
        *,
        limit: int = 25,
        starting_after: str | None = None,
    ) -> StripeListResult:
        """List Stripe customers."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        return await self._list_resource(
            self._stripe_client.v1.customers.list_async,
            limit=limit,
            starting_after=starting_after,
        )

    async def list_products(
        self,
        *,
        limit: int = 25,
        starting_after: str | None = None,
    ) -> StripeListResult:
        """List Stripe products with expanded default price."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        return await self._list_resource(
            self._stripe_client.v1.products.list_async,
            limit=limit,
            starting_after=starting_after,
            expand=["data.default_price"],
        )

    async def list_subscriptions(
        self,
        *,
        limit: int = 25,
        starting_after: str | None = None,
    ) -> StripeListResult:
        """List Stripe subscriptions across all statuses."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        return await self._list_resource(
            self._stripe_client.v1.subscriptions.list_async,
            limit=limit,
            starting_after=starting_after,
            extra_params={"status": "all"},
        )

    async def list_invoices(
        self,
        *,
        limit: int = 25,
        starting_after: str | None = None,
    ) -> StripeListResult:
        """List Stripe invoices."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        return await self._list_resource(
            self._stripe_client.v1.invoices.list_async,
            limit=limit,
            starting_after=starting_after,
        )

    async def list_charges(
        self,
        *,
        limit: int = 25,
        starting_after: str | None = None,
    ) -> StripeListResult:
        """List Stripe charges."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        return await self._list_resource(
            self._stripe_client.v1.charges.list_async,
            limit=limit,
            starting_after=starting_after,
        )

    # ------------------------------------------------------------------
    # Connected Accounts
    # ------------------------------------------------------------------

    async def create_account(
        self,
        *,
        type: AccountType | str | None = None,
        country: str | None = None,
        email: str | None = None,
        metadata: dict[str, str] | None = None,
        capabilities: dict[str, dict[str, bool]] | None = None,
        business_type: str | None = None,
        business_profile: dict[str, Any] | None = None,
    ) -> Account:
        """Create a Stripe Connect account."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        if type is not None and type not in AccountType:
            raise ValueError(
                f"Invalid account type: {type}. Must be one of {list(AccountType)}."
            )

        params: dict[str, Any] = {}
        if type is not None:
            params["type"] = type
        if country is not None:
            params["country"] = country
        if email is not None:
            params["email"] = email
        if metadata is not None:
            params["metadata"] = metadata
        if capabilities is not None:
            params["capabilities"] = capabilities
        if business_type is not None:
            params["business_type"] = business_type
        if business_profile is not None:
            params["business_profile"] = business_profile

        response = await self._provider_call(
            self._stripe_client.v1.accounts.create_async, params
        )
        return self._normalize_account(self._to_raw(response))

    async def retrieve_account(self, account_id: str) -> Account:
        """Retrieve a Stripe Connect account."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()

        response = await self._provider_call(
            self._stripe_client.v1.accounts.retrieve_async, account_id
        )
        return self._normalize_account(self._to_raw(response))

    async def update_account(
        self,
        account_id: str,
        *,
        metadata: dict[str, str] | None = None,
        business_profile: dict[str, Any] | None = None,
    ) -> Account:
        """Update a Stripe Connect account."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()

        params: dict[str, Any] = {}
        if metadata is not None:
            params["metadata"] = metadata
        if business_profile is not None:
            params["business_profile"] = business_profile

        response = await self._provider_call(
            self._stripe_client.v1.accounts.update_async, account_id, params
        )
        return self._normalize_account(self._to_raw(response))

    async def delete_account(self, account_id: str) -> Account:
        """Delete a Stripe Connect account."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()

        response = await self._provider_call(
            self._stripe_client.v1.accounts.delete_async, account_id
        )
        return self._normalize_account(self._to_raw(response))

    async def create_account_link(
        self,
        *,
        account_id: str,
        refresh_url: str,
        return_url: str,
        type: AccountLinkType | str,
    ) -> AccountLink:
        """Create a Stripe account onboarding or update link."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        if type not in AccountLinkType:
            raise ValueError(
                f"Invalid account link type: {type}. "
                f"Must be one of {list(AccountLinkType)}."
            )

        response = await self._provider_call(
            self._stripe_client.v1.account_links.create_async,
            {
                "account": account_id,
                "refresh_url": refresh_url,
                "return_url": return_url,
                "type": type,
            },
        )
        return self._normalize_account_link(self._to_raw(response))

    async def list_accounts(
        self,
        *,
        limit: int = 25,
        starting_after: str | None = None,
    ) -> StripeListResult:
        """List Stripe Connect accounts."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        return await self._list_resource(
            self._stripe_client.v1.accounts.list_async,
            limit=limit,
            starting_after=starting_after,
        )

    # ------------------------------------------------------------------
    # Transfers
    # ------------------------------------------------------------------

    async def create_transfer(
        self,
        *,
        amount: int,
        currency: str,
        destination: str,
        description: str | None = None,
        metadata: dict[str, str] | None = None,
        transfer_group: str | None = None,
    ) -> Transfer:
        """Create a Stripe transfer to a connected account."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        if not isinstance(amount, int) or isinstance(amount, bool) or amount <= 0:
            raise ValueError("amount must be a positive integer.")

        params: dict[str, Any] = {
            "amount": amount,
            "currency": currency,
            "destination": destination,
        }
        if description is not None:
            params["description"] = description
        if metadata is not None:
            params["metadata"] = metadata
        if transfer_group is not None:
            params["transfer_group"] = transfer_group

        response = await self._provider_call(
            self._stripe_client.v1.transfers.create_async, params
        )
        return self._normalize_transfer(self._to_raw(response))

    async def retrieve_transfer(self, transfer_id: str) -> Transfer:
        """Retrieve a Stripe transfer."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()

        response = await self._provider_call(
            self._stripe_client.v1.transfers.retrieve_async, transfer_id
        )
        return self._normalize_transfer(self._to_raw(response))

    async def list_transfers(
        self,
        *,
        limit: int = 25,
        starting_after: str | None = None,
        destination: str | None = None,
        transfer_group: str | None = None,
    ) -> StripeListResult:
        """List Stripe transfers."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        extra: dict[str, Any] = {}
        if destination is not None:
            extra["destination"] = destination
        if transfer_group is not None:
            extra["transfer_group"] = transfer_group

        return await self._list_resource(
            self._stripe_client.v1.transfers.list_async,
            limit=limit,
            starting_after=starting_after,
            extra_params=extra or None,
        )

    # ------------------------------------------------------------------
    # Payment Intents
    # ------------------------------------------------------------------

    async def create_payment_intent(
        self,
        *,
        amount: int,
        currency: str,
        customer_id: str | None = None,
        metadata: dict[str, str] | None = None,
        description: str | None = None,
        payment_method_types: list[str] | None = None,
        capture_method: CaptureMethod | str | None = None,
        transfer_data: dict[str, Any] | None = None,
        application_fee_amount: int | None = None,
        on_behalf_of: str | None = None,
    ) -> PaymentIntent:
        """Create a Stripe payment intent."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        if not isinstance(amount, int) or isinstance(amount, bool) or amount <= 0:
            raise ValueError("amount must be a positive integer.")
        if capture_method is not None and capture_method not in CaptureMethod:
            raise ValueError(
                f"Invalid capture_method: {capture_method}. "
                f"Must be one of {list(CaptureMethod)}."
            )

        params: dict[str, Any] = {"amount": amount, "currency": currency}
        if customer_id is not None:
            params["customer"] = customer_id
        if metadata is not None:
            params["metadata"] = metadata
        if description is not None:
            params["description"] = description
        if payment_method_types is not None:
            params["payment_method_types"] = payment_method_types
        if capture_method is not None:
            params["capture_method"] = capture_method
        if transfer_data is not None:
            params["transfer_data"] = transfer_data
        if application_fee_amount is not None:
            params["application_fee_amount"] = application_fee_amount
        if on_behalf_of is not None:
            params["on_behalf_of"] = on_behalf_of

        response = await self._provider_call(
            self._stripe_client.v1.payment_intents.create_async, params
        )
        return self._normalize_payment_intent(self._to_raw(response))

    async def retrieve_payment_intent(
        self, payment_intent_id: str, *, expand: list[str] | None = None
    ) -> PaymentIntent:
        """Retrieve a Stripe payment intent."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()

        params: dict[str, Any] = {}
        if expand is not None:
            params["expand"] = expand

        response = await self._provider_call(
            self._stripe_client.v1.payment_intents.retrieve_async,
            payment_intent_id,
            params,
        )
        return self._normalize_payment_intent(self._to_raw(response))

    async def confirm_payment_intent(
        self, payment_intent_id: str, *, payment_method: str | None = None
    ) -> PaymentIntent:
        """Confirm a Stripe payment intent."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()

        params: dict[str, Any] = {}
        if payment_method is not None:
            params["payment_method"] = payment_method

        response = await self._provider_call(
            self._stripe_client.v1.payment_intents.confirm_async,
            payment_intent_id,
            params,
        )
        return self._normalize_payment_intent(self._to_raw(response))

    async def capture_payment_intent(
        self, payment_intent_id: str, *, amount_to_capture: int | None = None
    ) -> PaymentIntent:
        """Capture an authorized Stripe payment intent."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()

        params: dict[str, Any] = {}
        if amount_to_capture is not None:
            params["amount_to_capture"] = amount_to_capture

        response = await self._provider_call(
            self._stripe_client.v1.payment_intents.capture_async,
            payment_intent_id,
            params,
        )
        return self._normalize_payment_intent(self._to_raw(response))

    async def cancel_payment_intent(
        self,
        payment_intent_id: str,
        *,
        cancellation_reason: CancellationReason | str | None = None,
    ) -> PaymentIntent:
        """Cancel a Stripe payment intent."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        if (
            cancellation_reason is not None
            and cancellation_reason not in CancellationReason
        ):
            raise ValueError(
                f"Invalid cancellation_reason: {cancellation_reason}. "
                f"Must be one of {list(CancellationReason)}."
            )

        params: dict[str, Any] = {}
        if cancellation_reason is not None:
            params["cancellation_reason"] = cancellation_reason

        response = await self._provider_call(
            self._stripe_client.v1.payment_intents.cancel_async,
            payment_intent_id,
            params,
        )
        return self._normalize_payment_intent(self._to_raw(response))

    async def list_payment_intents(
        self,
        *,
        limit: int = 25,
        starting_after: str | None = None,
        customer_id: str | None = None,
    ) -> StripeListResult:
        """List Stripe payment intents."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        extra: dict[str, Any] = {}
        if customer_id is not None:
            extra["customer"] = customer_id

        return await self._list_resource(
            self._stripe_client.v1.payment_intents.list_async,
            limit=limit,
            starting_after=starting_after,
            extra_params=extra or None,
        )

    # ------------------------------------------------------------------
    # Refunds
    # ------------------------------------------------------------------

    async def create_refund(
        self,
        *,
        payment_intent_id: str | None = None,
        charge_id: str | None = None,
        amount: int | None = None,
        reason: RefundReason | str | None = None,
        metadata: dict[str, str] | None = None,
        reverse_transfer: bool | None = None,
        refund_application_fee: bool | None = None,
    ) -> Refund:
        """Create a Stripe refund."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        if payment_intent_id is None and charge_id is None:
            raise ValueError(
                "At least one of `payment_intent_id` or `charge_id` must be provided."
            )
        if reason is not None and reason not in RefundReason:
            raise ValueError(
                f"Invalid refund reason: {reason}. Must be one of {list(RefundReason)}."
            )

        params: dict[str, Any] = {}
        if payment_intent_id is not None:
            params["payment_intent"] = payment_intent_id
        if charge_id is not None:
            params["charge"] = charge_id
        if amount is not None:
            params["amount"] = amount
        if reason is not None:
            params["reason"] = reason
        if metadata is not None:
            params["metadata"] = metadata
        if reverse_transfer is not None:
            params["reverse_transfer"] = reverse_transfer
        if refund_application_fee is not None:
            params["refund_application_fee"] = refund_application_fee

        response = await self._provider_call(
            self._stripe_client.v1.refunds.create_async, params
        )
        return self._normalize_refund(self._to_raw(response))

    async def retrieve_refund(self, refund_id: str) -> Refund:
        """Retrieve a Stripe refund."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()

        response = await self._provider_call(
            self._stripe_client.v1.refunds.retrieve_async, refund_id
        )
        return self._normalize_refund(self._to_raw(response))

    async def list_refunds(
        self,
        *,
        limit: int = 25,
        starting_after: str | None = None,
        payment_intent_id: str | None = None,
        charge_id: str | None = None,
    ) -> StripeListResult:
        """List Stripe refunds."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        extra: dict[str, Any] = {}
        if payment_intent_id is not None:
            extra["payment_intent"] = payment_intent_id
        if charge_id is not None:
            extra["charge"] = charge_id

        return await self._list_resource(
            self._stripe_client.v1.refunds.list_async,
            limit=limit,
            starting_after=starting_after,
            extra_params=extra or None,
        )

    # ------------------------------------------------------------------
    # Payouts
    # ------------------------------------------------------------------

    async def create_payout(
        self,
        *,
        amount: int,
        currency: str,
        description: str | None = None,
        metadata: dict[str, str] | None = None,
        destination: str | None = None,
        method: PayoutMethod | str | None = None,
        stripe_account: str | None = None,
    ) -> Payout:
        """Create a Stripe payout."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        if not isinstance(amount, int) or isinstance(amount, bool) or amount <= 0:
            raise ValueError("amount must be a positive integer.")
        if method is not None and method not in PayoutMethod:
            raise ValueError(
                f"Invalid payout method: {method}. Must be one of {list(PayoutMethod)}."
            )

        params: dict[str, Any] = {"amount": amount, "currency": currency}
        if description is not None:
            params["description"] = description
        if metadata is not None:
            params["metadata"] = metadata
        if destination is not None:
            params["destination"] = destination
        if method is not None:
            params["method"] = method

        options = self._connect_options(stripe_account)
        args: list[Any] = [params]
        if options is not None:
            args.append(options)

        response = await self._provider_call(
            self._stripe_client.v1.payouts.create_async, *args
        )
        return self._normalize_payout(self._to_raw(response))

    async def retrieve_payout(
        self, payout_id: str, *, stripe_account: str | None = None
    ) -> Payout:
        """Retrieve a Stripe payout."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()

        options = self._connect_options(stripe_account)
        response = await self._provider_call(
            self._stripe_client.v1.payouts.retrieve_async,
            payout_id,
            None,
            options,
        )
        return self._normalize_payout(self._to_raw(response))

    async def cancel_payout(
        self, payout_id: str, *, stripe_account: str | None = None
    ) -> Payout:
        """Cancel a pending Stripe payout."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()

        options = self._connect_options(stripe_account)
        response = await self._provider_call(
            self._stripe_client.v1.payouts.cancel_async,
            payout_id,
            None,
            options,
        )
        return self._normalize_payout(self._to_raw(response))

    async def list_payouts(
        self,
        *,
        limit: int = 25,
        starting_after: str | None = None,
        stripe_account: str | None = None,
    ) -> StripeListResult:
        """List Stripe payouts."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()
        return await self._list_resource(
            self._stripe_client.v1.payouts.list_async,
            limit=limit,
            starting_after=starting_after,
            options=self._connect_options(stripe_account),
        )

    # ------------------------------------------------------------------
    # Balance
    # ------------------------------------------------------------------

    async def retrieve_balance(self, *, stripe_account: str | None = None) -> Balance:
        """Retrieve Stripe account balance."""
        if self._stripe_client is None:
            raise PaymentsNotConnectedError()

        options = self._connect_options(stripe_account)
        response = await self._provider_call(
            self._stripe_client.v1.balance.retrieve_async,
            None,
            options,
        )
        return self._normalize_balance(self._to_raw(response))

    # ------------------------------------------------------------------
    # Webhooks
    # ------------------------------------------------------------------

    async def verify_webhook_event(
        self,
        *,
        payload: bytes,
        signature: str,
        webhook_secret: str | None = None,
    ) -> WebhookEvent:
        """Verify and parse a Stripe webhook payload."""
        secret = webhook_secret or self._config.webhook_secret
        if not secret:
            raise ValueError(
                "webhook_secret is required. Set PaymentsConfig.webhook_secret "
                "or pass webhook_secret to verify_webhook_event()."
            )

        try:
            event = stripe.Webhook.construct_event(
                payload=payload,
                sig_header=signature,
                secret=secret,
            )
        except Exception as exc:
            raise WebhookSignatureError(
                str(exc) or "Invalid webhook signature"
            ) from exc

        return self._normalize_event(self._to_raw(event))



"""Drop command - remove migration files."""

from __future__ import annotations

import shutil
from pathlib import Path
from typing import Annotated

import typer

from derp.config import ConfigError, DerpConfig
from derp.orm.migrations.journal import (
    get_migration_folders,
    load_journal,
    save_journal,
)


def drop(
    version: Annotated[
        str | None, typer.Argument(help="Version to drop (e.g., '0001')")
    ] = None,
    all_migrations: Annotated[
        bool, typer.Option("--all", help="Drop all migrations")
    ] = False,
    force: Annotated[
        bool, typer.Option("--force", "-f", help="Skip confirmation")
    ] = False,
) -> None:
    """Drop migration files (does not affect database).

    Removes migration folders and updates the journal. This does NOT
    reverse any changes in the database - it only removes the files.

    Use this to:
    - Remove a migration you haven't applied yet
    - Clean up and regenerate migrations during development
    """
    try:
        config = DerpConfig.load()
    except ConfigError as exc:
        typer.echo(f"Error: {exc}", err=True)
        raise typer.Exit(1)

    migrations_dir = Path(config.database.migrations_dir)

    if not migrations_dir.exists():
        typer.echo("No migrations directory found.")
        return

    journal = load_journal(migrations_dir)
    folders = get_migration_folders(migrations_dir)

    if not folders:
        typer.echo("No migration folders found.")
        return

    if all_migrations:
        _drop_all(migrations_dir, journal, folders, force)
    elif version:
        _drop_version(migrations_dir, journal, folders, version, force)
    else:
        # Interactive selection
        _drop_interactive(migrations_dir, journal, folders, force)


def _drop_all(migrations_dir, journal, folders, force: bool) -> None:
    """Drop all migrations."""
    typer.echo(f"Found {len(folders)} migration(s):")
    for version, folder in folders:
        typer.echo(f"  - {folder.name}")

    if not force:
        typer.echo("")
        typer.echo("This will delete all migration files and reset the journal.")
        typer.echo("The database will NOT be affected.")
        if not typer.confirm("Are you sure?"):
            raise typer.Abort()

    # Remove all folders
    for version, folder in folders:
        shutil.rmtree(folder)
        typer.echo(f"Deleted: {folder.name}")

    # Clear journal
    journal.entries = []
    save_journal(journal, migrations_dir)

    typer.echo("")
    typer.echo("All migrations dropped. Journal reset.")


def _drop_version(migrations_dir, journal, folders, version: str, force: bool) -> None:
    """Drop a specific version."""
    # Find folder
    target_folder = None
    for v, folder in folders:
        if v == version:
            target_folder = folder
            break

    if not target_folder:
        typer.echo(f"Migration version '{version}' not found.", err=True)
        typer.echo("Available versions:")
        for v, folder in folders:
            typer.echo(f"  - {v} ({folder.name})")
        raise typer.Exit(1)

    if not force:
        typer.echo(f"Will delete: {target_folder.name}")
        typer.echo("The database will NOT be affected.")
        if not typer.confirm("Are you sure?"):
            raise typer.Abort()

    # Remove folder
    shutil.rmtree(target_folder)
    typer.echo(f"Deleted: {target_folder.name}")

    # Update journal
    journal.remove_entry(version)
    save_journal(journal, migrations_dir)

    typer.echo("Migration dropped. Journal updated.")


def _drop_interactive(migrations_dir, journal, folders, force: bool) -> None:
    """Interactive migration selection."""
    typer.echo("Select migration to drop:")
    typer.echo("")

    for i, (version, folder) in enumerate(folders, 1):
        entry = journal.get_entry(version)
        tag = entry.tag if entry else "unknown"
        typer.echo(f"  {i}. {version} - {tag}")

    typer.echo(f"  {len(folders) + 1}. Drop all")
    typer.echo("  0. Cancel")
    typer.echo("")

    choice = typer.prompt("Enter number", type=int)

    if choice == 0:
        raise typer.Abort()
    elif choice == len(folders) + 1:
        _drop_all(migrations_dir, journal, folders, force)
    elif 1 <= choice <= len(folders):
        version, folder = folders[choice - 1]
        _drop_version(migrations_dir, journal, folders, version, force)
    else:
        typer.echo("Invalid choice.", err=True)
        raise typer.Exit(1)



"""Valkey-backed KV client using Valkey GLIDE."""

from __future__ import annotations

import asyncio
import math
from collections.abc import AsyncIterator, Sequence

from etils import epy

from derp.config import ValkeyConfig, ValkeyMode
from derp.kv.base import KVClient

with epy.lazy_imports():
    import glide


class ValkeyClient(KVClient):
    """Byte-level KV client backed by Valkey GLIDE."""

    supports_ttl = True
    supports_scan = True
    supports_batch = True

    def __init__(self, config: ValkeyConfig):
        addresses = [
            glide.NodeAddress(host=host, port=port) for host, port in config.addresses
        ]

        credentials: glide.ServerCredentials | None = (
            glide.ServerCredentials(username=config.username, password=config.password)
            if config.password is not None
            else None
        )

        self._is_cluster = config.mode == ValkeyMode.CLUSTER
        self._config: ValkeyConfig = config

        if self._is_cluster:
            self._glide_config: (
                glide.GlideClientConfiguration | glide.GlideClusterClientConfiguration
            ) = glide.GlideClusterClientConfiguration(
                addresses,
                credentials=credentials,
                use_tls=config.use_tls,
            )
        else:
            self._glide_config = glide.GlideClientConfiguration(
                addresses,
                credentials=credentials,
                use_tls=config.use_tls,
            )

        self._client: glide.GlideClient | glide.GlideClusterClient | None = None

    async def connect(self) -> None:
        if self._client is not None:
            return
        if self._is_cluster:
            self._client = await glide.GlideClusterClient.create(self._glide_config)
        else:
            self._client = await glide.GlideClient.create(self._glide_config)

    async def disconnect(self) -> None:
        if self._client is not None:
            await self._client.close()
            self._client = None

    @property
    def client(self) -> glide.GlideClient | glide.GlideClusterClient:
        if self._client is None:
            raise RuntimeError("Valkey client not connected. Call connect() first.")
        return self._client

    async def get(self, key: bytes) -> bytes | None:
        return await self.client.get(key)

    async def set(self, key: bytes, value: bytes, *, ttl: float | None = None) -> None:
        expiry = (
            glide.ExpirySet(glide.ExpiryType.SEC, math.ceil(ttl))
            if ttl is not None
            else None
        )
        await self.client.set(key, value, expiry=expiry)

    async def set_nx(
        self, key: bytes, value: bytes, *, ttl: float | None = None
    ) -> bool:
        expiry = (
            glide.ExpirySet(glide.ExpiryType.SEC, math.ceil(ttl))
            if ttl is not None
            else None
        )
        result = await self.client.set(
            key,
            value,
            conditional_set=glide.ConditionalChange.ONLY_IF_DOES_NOT_EXIST,
            expiry=expiry,
        )
        return result is not None

    async def delete(self, key: bytes) -> bool:
        return (await self.client.delete([key])) > 0

    async def exists(self, key: bytes) -> bool:
        return (await self.client.exists([key])) > 0

    async def mget(self, keys: Sequence[bytes]) -> Sequence[bytes | None]:
        if not keys:
            return []
        return await self.client.mget(list(keys))

    async def mset(
        self, items: Sequence[tuple[bytes, bytes]], *, ttl: float | None = None
    ) -> None:
        if not items:
            return
        mapping = {key: value for key, value in items}
        await self.client.mset(mapping)
        if ttl is not None:
            ttl_seconds = math.ceil(ttl)
            await asyncio.gather(
                *(self.client.expire(key, ttl_seconds) for key, _ in items)
            )

    async def delete_many(self, keys: Sequence[bytes]) -> int:
        if not keys:
            return 0
        return int(await self.client.delete(list(keys)))

    async def ttl(self, key: bytes) -> float | None:
        ttl = await self.client.ttl(key)
        if ttl is None or ttl < 0:
            return None
        return float(ttl)

    async def expire(self, key: bytes, ttl: float) -> bool:
        return bool(await self.client.expire(key, math.ceil(ttl)))

    async def scan(
        self, *, prefix: bytes | None = None, limit: int | None = None
    ) -> AsyncIterator[bytes]:
        match = prefix + b"*" if prefix else None
        count = 0

        if self._is_cluster:
            cluster_client: glide.GlideClusterClient = self.client  # type: ignore[assignment]
            cursor = glide.ClusterScanCursor()
            while not cursor.is_finished():
                result = await cluster_client.scan(cursor, match=match)
                cursor = result[0]
                keys: list[bytes] = result[1]  # type: ignore[assignment]
                for key in keys:
                    yield key
                    count += 1
                    if limit is not None and count >= limit:
                        return
        else:
            standalone_client: glide.GlideClient = self.client  # type: ignore[assignment]
            raw_cursor: bytes = b"0"
            while True:
                sa_result = await standalone_client.scan(raw_cursor, match=match)
                raw_cursor = sa_result[0]  # type: ignore[assignment]
                sa_keys: list[bytes] = sa_result[1]  # type: ignore[assignment]
                for key in sa_keys:
                    yield key
                    count += 1
                    if limit is not None and count >= limit:
                        return
                if raw_cursor == b"0":
                    break



"""Derp Studio FastAPI application factory."""

from __future__ import annotations

import asyncio
import json
import time
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from datetime import datetime
from datetime import time as dt_time
from decimal import Decimal
from pathlib import Path
from typing import Any
from uuid import UUID

from fastapi import Depends, FastAPI, HTTPException, Request
from fastapi.responses import FileResponse, PlainTextResponse, Response
from fastapi.staticfiles import StaticFiles

from derp.config import DerpConfig
from derp.derp_client import DerpClient
from derp.orm.migrations.introspect.postgres import PostgresIntrospector

ARTIFICIAL_LATENCY = 0.2
_INT_TYPES = frozenset({"integer", "bigint", "smallint", "int2", "int4", "int8"})
_FLOAT_TYPES = frozenset(
    {"real", "double precision", "float4", "float8", "numeric", "decimal"}
)
_BOOL_TYPES = frozenset({"boolean", "bool"})
_UUID_TYPES = frozenset({"uuid"})


_JSON_TYPES = frozenset({"json", "jsonb"})


def _coerce_value(value: object, col_type: str) -> object:
    """Coerce a JSON-decoded value to the Python type asyncpg expects."""
    if value is None:
        return None

    col_lower = col_type.lower().strip()

    # asyncpg expects JSON/JSONB as a serialised string.
    if col_lower in _JSON_TYPES:
        return json.dumps(value) if not isinstance(value, str) else value

    if not isinstance(value, str):
        return value

    col_lower = col_type.lower().strip()
    if col_lower == "date":
        # Value may be a full ISO timestamp; extract date part.
        return datetime.fromisoformat(value.replace("Z", "+00:00")).date()
    if col_lower.startswith("timestamp"):
        return datetime.fromisoformat(value.replace("Z", "+00:00"))
    if col_lower.startswith("time"):
        return dt_time.fromisoformat(value.replace("Z", "+00:00"))
    if col_lower in _INT_TYPES:
        return int(value)
    if col_lower in _FLOAT_TYPES or col_lower.startswith("numeric"):
        return Decimal(value)
    if col_lower in _BOOL_TYPES:
        return value.lower() in {"true", "t", "1", "yes"}
    if col_lower in _UUID_TYPES:
        return UUID(value)
    return value


def get_derp(request: Request) -> DerpClient:
    """Return the `DerpClient` from the request app state."""
    return request.app.state.derp_client


def _default_static_dir() -> Path:
    return Path(__file__).resolve().parent / "static"


def _missing_build_response(static_dir: Path) -> PlainTextResponse:
    message = (
        "Derp Studio frontend build is missing. "
        f"Expected: {static_dir / 'index.html'}. "
        f"Run `./scripts/build_studio.sh` from the project root."
    )
    return PlainTextResponse(message, status_code=503)


def _is_spa_path(path: str) -> bool:
    if path in {"", "api", "static"}:
        return False
    if path.startswith("api/") or path.startswith("static/"):
        return False
    return "." not in Path(path).name


def _serve_index(static_dir: Path) -> Response:
    index_path = static_dir / "index.html"
    if not index_path.exists():
        return _missing_build_response(static_dir)
    return FileResponse(index_path)


def create_app(
    *, static_dir: Path | None = None, enable_lifespan: bool = True
) -> FastAPI:
    """Create the Derp Studio FastAPI app."""

    @asynccontextmanager
    async def lifespan(app: FastAPI) -> AsyncIterator[None]:
        config = DerpConfig.load()
        derp_client = DerpClient(config)

        app.state.derp_client = derp_client

        try:
            await derp_client.connect()
        except Exception as exc:
            raise RuntimeError(
                "Failed to connect DerpClient during Studio startup."
            ) from exc

        yield

        await derp_client.disconnect()

    studio_static_dir = static_dir or _default_static_dir()

    app = FastAPI(
        title="Derp Studio",
        description="Derp Studio web interface",
        version="0.1.0",
        lifespan=lifespan if enable_lifespan else None,
    )

    @app.middleware("http")
    async def minimum_latency(request: Request, call_next):
        if ARTIFICIAL_LATENCY is not None:
            start = time.perf_counter()
            response = await call_next(request)
            elapsed = time.perf_counter() - start
            remaining = ARTIFICIAL_LATENCY - elapsed
            if remaining > 0:
                await asyncio.sleep(remaining)
            return response
        return await call_next(request)

    app.mount(
        "/static",
        StaticFiles(directory=str(studio_static_dir), check_dir=False),
        name="studio-static",
    )

    @app.get("/", include_in_schema=False)
    async def index() -> Response:
        return _serve_index(studio_static_dir)

    @app.get("/api/config")
    async def get_config(derp: DerpClient = Depends(get_derp)) -> dict:
        """Return loaded Derp configuration."""
        return derp.config.redacted_dump()

    @app.get("/api/storage/buckets")
    async def list_buckets(derp: DerpClient = Depends(get_derp)) -> dict:
        """List all S3 buckets."""
        if derp._storage is None:
            raise HTTPException(status_code=400, detail="Storage is not configured.")
        return {"buckets": await derp.storage.list_buckets()}

    @app.get("/api/storage/buckets/{bucket}/objects")
    async def list_objects(
        bucket: str,
        prefix: str = "",
        derp: DerpClient = Depends(get_derp),
    ) -> dict:
        """List objects in a bucket with prefix-based folder navigation."""
        if derp._storage is None:
            raise HTTPException(status_code=400, detail="Storage is not configured.")
        return await derp.storage.list_objects(bucket=bucket, prefix=prefix)

    @app.get("/api/storage/buckets/{bucket}/objects/info")
    async def get_object_info(
        bucket: str,
        key: str,
        derp: DerpClient = Depends(get_derp),
    ) -> dict:
        """Get object metadata (content type, size, etc.)."""
        if derp._storage is None:
            raise HTTPException(status_code=400, detail="Storage is not configured.")
        try:
            info = await derp.storage.head_object(bucket=bucket, key=key)
        except Exception as exc:
            raise HTTPException(
                status_code=404, detail=f"Object not found: {exc}"
            ) from exc
        return info

    @app.get("/api/storage/buckets/{bucket}/objects/content")
    async def get_object_content(
        bucket: str,
        key: str,
        derp: DerpClient = Depends(get_derp),
    ) -> Response:
        """Stream object content with its original content type."""
        if derp._storage is None:
            raise HTTPException(status_code=400, detail="Storage is not configured.")
        try:
            info = await derp.storage.head_object(bucket=bucket, key=key)
            data = await derp.storage.fetch_file(bucket=bucket, key=key)
        except Exception as exc:
            raise HTTPException(
                status_code=404, detail=f"Object not found: {exc}"
            ) from exc
        return Response(
            content=data,
            media_type=info["content_type"],
            headers={
                "Content-Length": str(len(data)),
                "Cache-Control": "private, max-age=60",
            },
        )

    @app.get("/api/email/templates")
    async def list_email_templates(derp: DerpClient = Depends(get_derp)) -> dict:
        """Return rendered previews for all configured email templates."""
        if derp._email is None or derp.config.email is None:
            raise HTTPException(status_code=400, detail="Email is not configured.")

        templates: list[dict[str, str]] = []
        for template_name, template in sorted(derp.email._templates.items()):
            if template_name == "base.html":
                continue

            content = derp.email._sources[template_name]
            if derp.email._base_template is not None:
                content = derp.email._base_template.render(
                    subject="<Email Subject>", content=content
                )
            templates.append({"name": template_name, "html": content})
        return {"templates": templates}

    # --- Database ---

    @app.get("/api/database/tables")
    async def list_tables(derp: DerpClient = Depends(get_derp)) -> dict:
        """List all database tables with column info and row counts."""
        introspector = PostgresIntrospector(derp.db.pool)
        snapshot = await introspector.introspect(
            schemas=derp.config.database.introspect_schemas,
            exclude_tables=derp.config.database.introspect_exclude_tables,
        )
        tables = []
        for table in snapshot.tables.values():
            columns = [
                {
                    "name": col.name,
                    "type": col.type,
                    "not_null": col.not_null,
                    "primary_key": col.primary_key,
                    "unique": col.unique,
                    "default": col.default,
                    "generated": col.generated or None,
                    "nullable": not col.not_null,
                }
                for col in table.columns.values()
            ]
            indexes = [
                {
                    "name": idx.name,
                    "columns": idx.columns,
                    "unique": idx.unique,
                    "method": idx.method,
                    "where": idx.where,
                }
                for idx in table.indexes.values()
            ]
            foreign_keys = [
                {
                    "name": fk.name,
                    "columns": fk.columns,
                    "references_table": fk.references_table,
                    "references_columns": fk.references_columns,
                    "references_schema": fk.references_schema,
                    "on_delete": fk.on_delete,
                    "on_update": fk.on_update,
                }
                for fk in table.foreign_keys.values()
            ]
            unique_constraints = [
                {
                    "name": uc.name,
                    "columns": uc.columns,
                }
                for uc in table.unique_constraints.values()
            ]
            check_constraints = [
                {
                    "name": cc.name,
                    "expression": cc.expression,
                }
                for cc in table.check_constraints.values()
            ]
            primary_key = (
                {"name": table.primary_key.name, "columns": table.primary_key.columns}
                if table.primary_key
                else None
            )
            qualified = (
                f'"{table.schema_name}"."{table.name}"'
                if table.schema_name != "public"
                else f'"{table.name}"'
            )
            row_count = await derp.db.table(qualified).select("*").count()
            tables.append(
                {
                    "name": table.name,
                    "schema": table.schema_name,
                    "columns": columns,
                    "row_count": row_count,
                    "indexes": indexes,
                    "foreign_keys": foreign_keys,
                    "unique_constraints": unique_constraints,
                    "check_constraints": check_constraints,
                    "primary_key": primary_key,
                }
            )
        return {"tables": tables}

    @app.get("/api/database/tables/{table}/rows")
    async def list_table_rows(
        table: str,
        limit: int = 50,
        offset: int = 0,
        schema: str = "public",
        derp: DerpClient = Depends(get_derp),
    ) -> dict:
        """Fetch rows from a table with pagination."""
        limit = min(max(limit, 1), 500)
        offset = max(offset, 0)

        qualified = f'"{schema}"."{table}"' if schema != "public" else f'"{table}"'
        total = await derp.db.table(qualified).select("*").count()
        rows = (
            await derp.db.table(qualified)
            .select("*")
            .limit(limit)
            .offset(offset)
            .execute()
        )
        return {"rows": rows, "total": total, "limit": limit, "offset": offset}

    @app.post("/api/database/tables/{table}/delete-rows")
    async def delete_table_rows(
        table: str,
        request: Request,
        schema: str = "public",
        derp: DerpClient = Depends(get_derp),
    ) -> dict:
        """Delete rows from a table by primary key values."""
        body = await request.json()
        row_keys: list[dict] = body.get("rows", [])
        if not row_keys:
            raise HTTPException(status_code=400, detail="No rows specified")

        # Resolve primary key columns from introspection
        introspector = PostgresIntrospector(derp.db.pool)
        snapshot = await introspector.introspect(
            schemas=derp.config.database.introspect_schemas,
            exclude_tables=derp.config.database.introspect_exclude_tables,
        )
        table_key = f"{schema}.{table}" if schema != "public" else table
        table_snapshot = snapshot.tables.get(table_key)
        if table_snapshot is None:
            raise HTTPException(status_code=404, detail=f"Table {table!r} not found")

        pk_columns = [
            col.name for col in table_snapshot.columns.values() if col.primary_key
        ]
        if not pk_columns:
            raise HTTPException(
                status_code=400,
                detail=f"Table {table!r} has no primary key",
            )

        qualified = f'"{schema}"."{table}"' if schema != "public" else f'"{table}"'

        deleted = 0
        for row_key in row_keys:
            q = derp.db.table(qualified).delete()
            for pk_col in pk_columns:
                if pk_col not in row_key:
                    raise HTTPException(
                        status_code=400,
                        detail=f"Missing primary key column {pk_col!r}",
                    )
                q = q.eq(pk_col, row_key[pk_col])
            result = await q.returning("*").execute()
            deleted += len(result)

        return {"deleted": deleted}

    @app.post("/api/database/tables/{table}/update-row")
    async def update_table_row(
        table: str,
        request: Request,
        schema: str = "public",
        derp: DerpClient = Depends(get_derp),
    ) -> dict:
        """Update a single row by primary key values."""
        body = await request.json()
        key: dict = body.get("key", {})
        values: dict = body.get("values", {})
        if not key:
            raise HTTPException(status_code=400, detail="No key specified")
        if not values:
            raise HTTPException(status_code=400, detail="No values specified")

        introspector = PostgresIntrospector(derp.db.pool)
        snapshot = await introspector.introspect(
            schemas=derp.config.database.introspect_schemas,
            exclude_tables=derp.config.database.introspect_exclude_tables,
        )
        table_key = f"{schema}.{table}" if schema != "public" else table
        table_snapshot = snapshot.tables.get(table_key)
        if table_snapshot is None:
            raise HTTPException(status_code=404, detail=f"Table {table!r} not found")

        pk_columns = [
            col.name for col in table_snapshot.columns.values() if col.primary_key
        ]
        if not pk_columns:
            raise HTTPException(
                status_code=400,
                detail=f"Table {table!r} has no primary key",
            )

        qualified = f'"{schema}"."{table}"' if schema != "public" else f'"{table}"'

        col_types = {c.name: c.type for c in table_snapshot.columns.values()}

        coerced = {
            col: _coerce_value(val, col_types.get(col, "text"))
            for col, val in values.items()
        }

        q = derp.db.table(qualified).update(coerced)
        for pk_col in pk_columns:
            if pk_col not in key:
                raise HTTPException(
                    status_code=400,
                    detail=f"Missing primary key column {pk_col!r}",
                )
            q = q.eq(pk_col, key[pk_col])

        result = await q.returning("*").execute()
        updated = len(result)

        return {"updated": updated}

    # --- KV ---

    @app.get("/api/kv/keys")
    async def list_kv_keys(
        prefix: str = "",
        limit: int = 100,
        derp: DerpClient = Depends(get_derp),
    ) -> dict:
        """Scan KV keys with optional prefix filtering."""
        if derp._kv is None:
            raise HTTPException(status_code=400, detail="KV is not configured.")
        limit = min(max(limit, 1), 1000)
        prefix_bytes = prefix.encode() if prefix else None
        keys: list[str] = []
        async for key in derp.kv.scan(prefix=prefix_bytes, limit=limit):
            keys.append(key.decode("utf-8", errors="replace"))
        return {"keys": keys}

    @app.get("/api/kv/keys/info")
    async def get_kv_key_info(
        key: str,
        derp: DerpClient = Depends(get_derp),
    ) -> dict:
        """Get value and TTL for a single KV key."""
        if derp._kv is None:
            raise HTTPException(status_code=400, detail="KV is not configured.")
        key_bytes = key.encode()
        value = await derp.kv.get(key_bytes)
        if value is None:
            raise HTTPException(status_code=404, detail=f"Key not found: {key}")
        ttl = await derp.kv.ttl(key_bytes)
        return {
            "key": key,
            "value": value.decode("utf-8", errors="replace"),
            "ttl": ttl,
            "size": len(value),
        }

    @app.delete("/api/kv/keys")
    async def delete_kv_key(
        request: Request,
        derp: DerpClient = Depends(get_derp),
    ) -> dict:
        """Delete a single KV key."""
        if derp._kv is None:
            raise HTTPException(status_code=400, detail="KV is not configured.")
        body: dict[str, Any] = await request.json()
        key: str = body.get("key", "")
        if not key:
            raise HTTPException(status_code=400, detail="No key specified.")
        deleted = await derp.kv.delete(key.encode())
        return {"deleted": deleted}

    # --- Auth ---

    @app.get("/api/auth/users")
    async def list_auth_users(
        limit: int = 100,
        derp: DerpClient = Depends(get_derp),
    ) -> dict:
        """List auth users."""
        if derp._auth is None:
            raise HTTPException(status_code=400, detail="Auth is not configured.")
        limit = min(max(limit, 1), 500)
        users = await derp.auth.list_users(limit=limit)
        return {"users": users}

    @app.get("/api/auth/sessions")
    async def list_auth_sessions(
        limit: int = 100,
        derp: DerpClient = Depends(get_derp),
    ) -> dict:
        """List auth sessions."""
        if derp._auth is None:
            raise HTTPException(status_code=400, detail="Auth is not configured.")
        limit = min(max(limit, 1), 500)
        sessions = await derp.auth.list_sessions(limit=limit)
        return {"sessions": sessions}

    # --- Payments ---

    @app.get("/api/payments/customers")
    async def list_customers(
        limit: int = 25,
        starting_after: str | None = None,
        derp: DerpClient = Depends(get_derp),
    ) -> dict:
        """List Stripe customers."""
        if derp._payments is None:
            raise HTTPException(status_code=400, detail="Payments is not configured.")
        result = await derp.payments.list_customers(
            limit=limit, starting_after=starting_after
        )
        return {"data": result.data, "has_more": result.has_more}

    @app.get("/api/payments/products")
    async def list_products(
        limit: int = 25,
        starting_after: str | None = None,
        derp: DerpClient = Depends(get_derp),
    ) -> dict:
        """List Stripe products."""
        if derp._payments is None:
            raise HTTPException(status_code=400, detail="Payments is not configured.")
        result = await derp.payments.list_products(
            limit=limit, starting_after=starting_after
        )
        return {"data": result.data, "has_more": result.has_more}

    @app.get("/api/payments/subscriptions")
    async def list_subscriptions(
        limit: int = 25,
        starting_after: str | None = None,
        derp: DerpClient = Depends(get_derp),
    ) -> dict:
        """List Stripe subscriptions."""
        if derp._payments is None:
            raise HTTPException(status_code=400, detail="Payments is not configured.")
        result = await derp.payments.list_subscriptions(
            limit=limit, starting_after=starting_after
        )
        return {"data": result.data, "has_more": result.has_more}

    @app.get("/api/payments/invoices")
    async def list_invoices(
        limit: int = 25,
        starting_after: str | None = None,
        derp: DerpClient = Depends(get_derp),
    ) -> dict:
        """List Stripe invoices."""
        if derp._payments is None:
            raise HTTPException(status_code=400, detail="Payments is not configured.")
        result = await derp.payments.list_invoices(
            limit=limit, starting_after=starting_after
        )
        return {"data": result.data, "has_more": result.has_more}

    @app.get("/api/payments/charges")
    async def list_charges(
        limit: int = 25,
        starting_after: str | None = None,
        derp: DerpClient = Depends(get_derp),
    ) -> dict:
        """List Stripe charges."""
        if derp._payments is None:
            raise HTTPException(status_code=400, detail="Payments is not configured.")
        result = await derp.payments.list_charges(
            limit=limit, starting_after=starting_after
        )
        return {"data": result.data, "has_more": result.has_more}

    @app.get("/{path:path}", include_in_schema=False)
    async def spa_fallback(path: str) -> Response:
        if not _is_spa_path(path):
            raise HTTPException(status_code=404, detail="Not Found")
        return _serve_index(studio_static_dir)

    return app



"""Table base class for Derp ORM using Pydantic."""

from __future__ import annotations

from typing import Any, ClassVar

from pydantic import BaseModel

from derp.orm.fields import FieldInfo


class ColumnAccessor:
    """Provides attribute access to table columns for query building."""

    def __init__(self, columns: dict[str, FieldInfo[Any]]):
        self._columns = columns

    def __getattr__(self, name: str) -> FieldInfo[Any]:
        if name.startswith("_"):
            raise AttributeError(name)
        if name not in self._columns:
            raise AttributeError(f"Column '{name}' not found")
        return self._columns[name]

    def __iter__(self):
        return iter(self._columns.values())

    def __contains__(self, item: str) -> bool:
        return item in self._columns


class TableMeta(type(BaseModel)):
    """Metaclass for Table that processes field definitions."""

    def __new__(
        mcs,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        table: str | None = None,
        **kwargs: Any,
    ) -> Any:
        # Store table name in class namespace
        if table is not None:
            namespace["__table_name__"] = table
            namespace["__explicit_table__"] = True

            # Enforce: if parent has __explicit_table__, child must use same table name
            for base in bases:
                parent_table = getattr(base, "__table_name__", None)
                parent_explicit = getattr(base, "__explicit_table__", False)
                if (
                    parent_explicit
                    and parent_table is not None
                    and table != parent_table
                ):
                    raise TypeError(
                        f"Table '{name}' uses table name '{table}' but its parent "
                        f"'{base.__name__}' uses '{parent_table}'. Inherited tables "
                        f"must use the same table name as their parent."
                    )
        elif "__table_name__" not in namespace:
            # Default to lowercase class name
            namespace["__table_name__"] = name.lower()
            namespace["__explicit_table__"] = False

        # Extract FieldInfo objects BEFORE Pydantic processes them
        field_infos: dict[str, FieldInfo[Any]] = {}
        annotations = namespace.get("__annotations__", {})

        for field_name, type_hint in annotations.items():
            if field_name.startswith("_"):
                continue

            default = namespace.get(field_name)
            if isinstance(default, FieldInfo):
                field_infos[field_name] = default
                # Remove from namespace so Pydantic doesn't try to process it
                # Pydantic needs to see either no default or a proper Pydantic default
                del namespace[field_name]

        # Store field infos for processing after class creation
        namespace["__derp_field_infos__"] = field_infos

        # Create the class (Pydantic will process it)
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)

        # Process field definitions after class creation
        if name != "Table":  # Skip base Table class
            mcs._process_fields(cls)

        return cls

    @staticmethod
    def _process_fields(cls: Any) -> None:
        """Process FieldInfo annotations and set up column metadata."""
        columns: dict[str, FieldInfo[Any]] = {}
        table_name = getattr(cls, "__table_name__", cls.__name__.lower())

        # Collect columns from parent classes (in MRO order, excluding Table and object)
        for base in reversed(cls.__mro__):
            if base is cls:
                continue
            base_columns = getattr(base, "__columns__", None)
            if base_columns is not None:
                for field_name, info in base_columns.items():
                    # Clone FieldInfo with this class's table name
                    field_info: FieldInfo[Any] = FieldInfo(
                        field_type=info.field_type,
                        primary_key=info.primary_key,
                        unique=info.unique,
                        nullable=info.nullable,
                        default=info.default,
                        foreign_key=info.foreign_key,
                        index=info.index,
                        _table_name=table_name,
                        _field_name=field_name,
                    )
                    columns[field_name] = field_info

        # Get the stored field infos for this class (may override parent fields)
        field_infos: dict[str, FieldInfo[Any]] = getattr(
            cls, "__derp_field_infos__", {}
        )

        for field_name, info in field_infos.items():
            # Clone FieldInfo with table/field name set
            field_info = FieldInfo(
                field_type=info.field_type,
                primary_key=info.primary_key,
                unique=info.unique,
                nullable=info.nullable,
                default=info.default,
                foreign_key=info.foreign_key,
                index=info.index,
                _table_name=table_name,
                _field_name=field_name,
            )
            columns[field_name] = field_info

        cls.__columns__ = columns
        cls.c = ColumnAccessor(columns)


class Table(BaseModel, metaclass=TableMeta):
    """Base class for all Derp table definitions.

    Example:
        class User(Table, table="users"):
            id: int = Field(Serial(), primary_key=True)
            name: str = Field(Varchar(255))
            email: str = Field(Varchar(255), unique=True)

        # Access columns via .c attribute for query building:
        db.select(User).where(User.c.id == 1)
    """

    __table_name__: ClassVar[str]
    __explicit_table__: ClassVar[bool]
    __columns__: ClassVar[dict[str, FieldInfo[Any]]]
    __indexes__: ClassVar[list[tuple[str, ...]]]
    c: ClassVar[ColumnAccessor]

    model_config = {"from_attributes": True}

    @classmethod
    def get_table_name(cls) -> str:
        """Get the SQL table name."""
        return cls.__table_name__

    @classmethod
    def get_columns(cls) -> dict[str, FieldInfo[Any]]:
        """Get all column definitions."""
        return getattr(cls, "__columns__", {})

    @classmethod
    def get_primary_key(cls) -> tuple[str, FieldInfo[Any]] | None:
        """Get the primary key column if any."""
        for name, info in cls.get_columns().items():
            if info.primary_key:
                return (name, info)
        return None

    @classmethod
    def to_ddl(cls) -> str:
        """Generate CREATE TABLE DDL statement."""
        columns = cls.get_columns()
        if not columns:
            raise ValueError(f"Table {cls.__name__} has no columns defined")

        table_name = cls.get_table_name()
        column_defs: list[str] = []
        constraints: list[str] = []
        indexes: list[str] = []

        for col_name, info in columns.items():
            col_def = f"    {col_name} {info.field_type.sql_type()}"

            if info.primary_key:
                col_def += " PRIMARY KEY"
            if not info.nullable and not info.primary_key:
                col_def += " NOT NULL"
            if info.unique and not info.primary_key:
                col_def += " UNIQUE"
            if info.default is not None:
                if isinstance(info.default, str) and (
                    info.default.endswith("()")
                    or info.default.upper() in ("CURRENT_TIMESTAMP", "TRUE", "FALSE")
                ):
                    col_def += f" DEFAULT {info.default}"
                elif isinstance(info.default, bool):
                    col_def += f" DEFAULT {str(info.default).upper()}"
                elif isinstance(info.default, int | float):
                    col_def += f" DEFAULT {info.default}"
                else:
                    escaped = str(info.default).replace("'", "''")
                    col_def += f" DEFAULT '{escaped}'"

            column_defs.append(col_def)

            # Foreign key constraints
            if info.foreign_key:
                fk = info.foreign_key
                ref_table, ref_col = fk.reference.split(".")
                constraint = (
                    f"    FOREIGN KEY ({col_name}) REFERENCES {ref_table}({ref_col})"
                )
                if fk.on_delete:
                    constraint += f" ON DELETE {fk.on_delete}"
                if fk.on_update:
                    constraint += f" ON UPDATE {fk.on_update}"
                constraints.append(constraint)

            # Indexes
            if info.index:
                indexes.append(
                    f"CREATE INDEX idx_{table_name}_{col_name} ON "
                    f"{table_name}({col_name});"
                )

        # Composite indexes from __indexes__
        composite = getattr(cls, "__indexes__", [])
        for cols in composite:
            idx_name = "idx_" + table_name + "_" + "_".join(cols)
            col_list = ", ".join(cols)
            indexes.append(f"CREATE INDEX {idx_name} ON {table_name}({col_list});")

        all_defs = column_defs + constraints
        ddl = f"CREATE TABLE {table_name} (\n"
        ddl += ",\n".join(all_defs)
        ddl += "\n);"

        if indexes:
            ddl += "\n\n" + "\n".join(indexes)

        return ddl


def get_column_ref(table: type[Table], column_name: str) -> FieldInfo[Any]:
    """Get a column reference for query building."""
    columns = table.get_columns()
    if column_name not in columns:
        raise ValueError(f"Column {column_name} not found in table {table.__name__}")
    return columns[column_name]



"""Base interface for KV clients."""

from __future__ import annotations

import abc
import asyncio
import json
from collections.abc import AsyncIterator, Awaitable, Callable, Sequence
from typing import Any


class KVClient(abc.ABC):
    """Byte-level async KV client."""

    supports_ttl: bool
    supports_scan: bool
    supports_batch: bool

    @abc.abstractmethod
    async def connect(self) -> None:
        """Connect to the store."""

    @abc.abstractmethod
    async def disconnect(self) -> None:
        """Disconnect from the store."""

    @abc.abstractmethod
    async def get(self, key: bytes) -> bytes | None:
        """Fetch a value by key."""

    @abc.abstractmethod
    async def set(self, key: bytes, value: bytes, *, ttl: float | None = None) -> None:
        """Set a value by key."""

    @abc.abstractmethod
    async def delete(self, key: bytes) -> bool:
        """Delete a key."""

    @abc.abstractmethod
    async def exists(self, key: bytes) -> bool:
        """Check if a key exists."""

    @abc.abstractmethod
    async def mget(self, keys: Sequence[bytes]) -> Sequence[bytes | None]:
        """Fetch multiple keys."""

    @abc.abstractmethod
    async def mset(
        self, items: Sequence[tuple[bytes, bytes]], *, ttl: float | None = None
    ) -> None:
        """Set multiple key/value pairs."""

    @abc.abstractmethod
    async def delete_many(self, keys: Sequence[bytes]) -> int:
        """Delete multiple keys."""

    @abc.abstractmethod
    async def ttl(self, key: bytes) -> float | None:
        """Return remaining TTL in seconds, or None."""

    @abc.abstractmethod
    async def expire(self, key: bytes, ttl: float) -> bool:
        """Set TTL for a key. Returns False if missing."""

    @abc.abstractmethod
    async def set_nx(
        self, key: bytes, value: bytes, *, ttl: float | None = None
    ) -> bool:
        """Set a value only if the key does not exist. Returns True if set."""

    @abc.abstractmethod
    async def scan(
        self, *, prefix: bytes | None = None, limit: int | None = None
    ) -> AsyncIterator[bytes]:
        """Iterate keys with optional prefix and limit."""
        yield b""  # pragma: no cover
        raise NotImplementedError  # pragma: no cover

    async def guarded_get(
        self,
        cache_key: bytes,
        *,
        compute: Callable[[], Awaitable[bytes]],
        ttl: float,
        lock_ttl: float = 2.0,
        retry_delay: float = 0.05,
    ) -> bytes:
        """Fetch from cache with stampede protection.

        On a cache miss, only one caller acquires a lock and computes the
        value. Other callers wait for the cache to be populated. If retries
        are exhausted, the caller falls through and computes directly.

        The wait budget is derived from ``lock_ttl`` so waiters keep
        retrying for the full duration the lock could be held.

        Args:
            cache_key: The cache key to read/write.
            compute: Async callable that produces the value on cache miss.
            ttl: TTL in seconds for the cached value.
            lock_ttl: TTL in seconds for the lock key.
            retry_delay: Seconds to sleep between retry attempts.

        Returns:
            The cached or freshly computed value.
        """
        cached = await self.get(cache_key)
        if cached is not None:
            return cached

        lock_key = cache_key + b":lock"
        acquired = await self.set_nx(lock_key, b"1", ttl=lock_ttl)

        if acquired:
            try:
                cached = await self.get(cache_key)
                if cached is not None:
                    return cached

                value = await compute()
                await self.set(cache_key, value, ttl=ttl)
                return value
            finally:
                await self.delete(lock_key)

        max_retries = round(lock_ttl / retry_delay)
        for _ in range(max_retries):
            await asyncio.sleep(retry_delay)
            cached = await self.get(cache_key)
            if cached is not None:
                return cached

        return await compute()

    async def idempotent_execute(
        self,
        *,
        key: str,
        compute: Callable[[], Awaitable[Any]],
        status_code: int = 200,
        ttl: float = 86400,
        key_prefix: str = "derp:idempotency",
    ) -> tuple[Any, int, bool]:
        """Execute idempotently: run ``compute`` once per key.

        On the first call for a given key, ``compute`` is invoked and
        the result is cached. Subsequent calls return the cached result
        without re-invoking ``compute``. Uses :meth:`guarded_get` for
        stampede protection.

        Args:
            key: Idempotency key (typically from a client header).
            compute: Async callable producing a JSON-serializable result.
            status_code: HTTP status code to cache alongside the body.
            ttl: Cache TTL in seconds (default 24h).
            key_prefix: KV key prefix.

        Returns:
            ``(body, status_code, is_replay)`` — *body* is the
            deserialized result, *status_code* is the cached status,
            and *is_replay* is ``True`` when the cached value was used.
        """
        cache_key = f"{key_prefix}:{key}".encode()
        was_computed = False

        async def _compute() -> bytes:
            nonlocal was_computed
            was_computed = True
            result = await compute()
            payload = json.dumps(
                {"status_code": status_code, "body": result},
                default=str,
            )
            return payload.encode()

        raw = await self.guarded_get(cache_key, compute=_compute, ttl=ttl)
        parsed = json.loads(raw)
        return parsed["body"], parsed["status_code"], not was_computed

    async def already_processed(
        self,
        *,
        event_id: str,
        ttl: float = 86400,
        key_prefix: str = "derp:webhook",
    ) -> bool:
        """Check if an event has already been processed.

        Uses :meth:`set_nx` to atomically mark the event. Returns
        ``True`` if the event was already seen, ``False`` on first call.

        Args:
            event_id: Unique event identifier (e.g. Stripe event ID).
            ttl: How long to remember the event (default 24h).
            key_prefix: KV key prefix.
        """
        cache_key = f"{key_prefix}:{event_id}".encode()
        acquired = await self.set_nx(cache_key, b"1", ttl=ttl)
        return not acquired



"""Core authentication service."""

from __future__ import annotations

import uuid
from datetime import UTC, datetime, timedelta
from typing import Any

from derp.auth.base import BaseAuthClient
from derp.auth.email import EmailClient
from derp.auth.exceptions import (
    ConfirmationTokenInvalidError,
    ConfirmationURLMissingError,
    EmailNotConfirmedError,
    InvalidCredentialsError,
    MagicLinkExpiredError,
    NotOrgMemberError,
    OrgAlreadyExistsError,
    OrgLastOwnerError,
    OrgMemberExistsError,
    OrgMemberNotFoundError,
    OrgNotFoundError,
    PasswordValidationError,
    RecoveryTokenInvalidError,
    RefreshTokenReusedError,
    RefreshTokenRevokedError,
    SessionExpiredError,
    SessionNotFoundError,
    SignupDisabledError,
    UserAlreadyExistsError,
    UserNotActiveError,
    UserNotFoundError,
)
from derp.auth.jwt import TokenPair, create_token_pair, decode_token
from derp.auth.models import (
    AuthOrganization,
    AuthOrgMember,
    AuthProvider,
    AuthRequest,
    AuthSession,
    AuthUser,
    OrgInfo,
    OrgMemberInfo,
    SessionInfo,
    UserInfo,
)
from derp.auth.password import (
    Argon2Hasher,
    PasswordHasher,
    generate_secure_token,
    validate_password,
)
from derp.auth.providers.base import BaseOAuthProvider
from derp.auth.providers.github import GitHubProvider
from derp.auth.providers.google import GoogleProvider
from derp.config import NativeAuthConfig
from derp.kv.base import KVClient
from derp.orm import DatabaseEngine


class NativeAuthClient(BaseAuthClient):
    """Native authentication client (email/password, magic link, OAuth)."""

    def __init__(self, config: NativeAuthConfig):
        self._config: NativeAuthConfig = config
        self._hasher: PasswordHasher = Argon2Hasher()
        self._email_client: EmailClient | None = None
        self._oauth_providers: dict[AuthProvider, BaseOAuthProvider] = {}
        self._database_client: DatabaseEngine | None = None
        self._kv_client: KVClient | None = None

        if self._config.google_oauth is not None:
            self._oauth_providers[AuthProvider.GOOGLE] = GoogleProvider(
                self._config.google_oauth
            )
        if self._config.github_oauth is not None:
            self._oauth_providers[AuthProvider.GITHUB] = GitHubProvider(
                self._config.github_oauth
            )

    def set_db(self, db: DatabaseEngine | None) -> None:
        """Set the database client."""
        self._database_client = db

    def _db(self) -> DatabaseEngine:
        """Get the database client."""
        if self._database_client is None:
            raise ValueError("Database client not set. Must call `set_db()` first.")
        return self._database_client

    def set_kv(self, kv: KVClient | None) -> None:
        """Set the KV store for caching and token storage."""
        self._kv_client = kv

    def _kv(self) -> KVClient:
        """Get the KV client. Required for token operations."""
        if self._kv_client is None:
            raise ValueError(
                "KV client not set. Token operations (recovery, confirmation, "
                "magic link) require a KV store. Call `set_kv()` first."
            )
        return self._kv_client

    def set_email(self, email_client: EmailClient | None) -> None:
        """Set the email client."""
        self._email_client = email_client

    def _email(self) -> EmailClient:
        """Get the email client."""
        if self._email_client is None:
            raise ValueError("Email client not set. Must call `set_email()` first.")
        return self._email_client

    async def _invalidate_user_cache(
        self, user_id: str | uuid.UUID, email: str | None = None
    ) -> None:
        """Invalidate cached user data in KV store."""
        if self._kv_client is not None:
            await self._kv_client.delete(
                f"{self._config.cache_prefix}:user:{user_id}".encode()
            )
            if email is not None:
                await self._kv_client.delete(
                    f"{self._config.cache_prefix}:user:email:{email.lower()}".encode()
                )

    # =========================================================================
    # User Management
    # =========================================================================

    def _to_user_info(self, user: AuthUser) -> UserInfo:
        """Convert an internal AuthUser ORM model to a public UserInfo."""
        return UserInfo(
            id=str(user.id),
            email=user.email,
            first_name=user.first_name,
            last_name=user.last_name,
            username=user.username,
            image_url=user.image_url,
            role=user.role,
            is_active=user.is_active,
            is_superuser=user.is_superuser,
            created_at=user.created_at,
            updated_at=user.updated_at,
            last_sign_in_at=user.last_sign_in_at,
            email_confirmed_at=user.email_confirmed_at,
            metadata={
                "provider": user.provider.value,
                "provider_id": user.provider_id,
            },
        )

    async def _fetch_user(self, user_id: str | uuid.UUID) -> AuthUser | None:
        """Fetch a user by ID (internal, with caching)."""
        if self._config.use_kv_cache and self._kv_client is not None:
            cache_key = f"{self._config.cache_prefix}:user:{user_id}".encode()

            async def _compute() -> bytes:
                row = await (
                    self._db()
                    .select(AuthUser)
                    .where(AuthUser.c.id == str(user_id))
                    .first_or_none()
                )
                if row is None:
                    return b""
                return row.model_dump_json().encode()

            cached = await self._kv_client.guarded_get(
                cache_key,
                compute=_compute,
                ttl=self._config.cache_user_ttl_seconds,
            )
            if cached == b"":
                return None
            return AuthUser.model_validate_json(cached)

        return await (
            self._db()
            .select(AuthUser)
            .where(AuthUser.c.id == str(user_id))
            .first_or_none()
        )

    async def get_user(self, user_id: str | uuid.UUID) -> UserInfo | None:
        """Get a user by their ID."""
        user = await self._fetch_user(user_id)
        return self._to_user_info(user) if user is not None else None

    async def list_users(
        self, *, limit: int | None = None, offset: int | None = None
    ) -> list[UserInfo]:
        """List users ordered by creation date (newest first)."""
        q = self._db().select(AuthUser).order_by(AuthUser.c.created_at, asc=False)
        if limit is not None:
            q = q.limit(limit)
        if offset is not None:
            q = q.offset(offset)
        return [self._to_user_info(u) for u in await q.execute()]

    async def _get_user_by_email(self, email: str) -> AuthUser | None:
        """Get a user by their email address (internal use only).

        Unlike ``get_user``, negative results (user not found) are **not**
        cached because email lookups are used in write paths (sign-up, OAuth)
        where a subsequent insert would leave a stale "not found" entry.
        """
        normalized = email.lower()

        if self._config.use_kv_cache and self._kv_client is not None:
            cache_key = f"{self._config.cache_prefix}:user:email:{normalized}".encode()
            cached = await self._kv_client.get(cache_key)
            if cached is not None:
                return AuthUser.model_validate_json(cached)

            user = await (
                self._db()
                .select(AuthUser)
                .where(AuthUser.c.email == normalized)
                .first_or_none()
            )
            if user is not None:
                await self._kv_client.set(
                    cache_key,
                    user.model_dump_json().encode(),
                    ttl=self._config.cache_user_ttl_seconds,
                )
            return user

        return await (
            self._db()
            .select(AuthUser)
            .where(AuthUser.c.email == normalized)
            .first_or_none()
        )

    async def update_user(
        self,
        *,
        user_id: str | uuid.UUID,
        email: str | None = None,
        **kwargs: Any,
    ) -> UserInfo:
        """Update user data."""
        user = await self._fetch_user(user_id=user_id)
        if not user:
            raise UserNotFoundError()

        updates: dict[str, Any] = {"updated_at": datetime.now(UTC)}

        if email is not None:
            updates["email"] = email.lower()

        for key, value in kwargs.items():
            if key in AuthUser.c:
                updates[key] = value
            else:
                raise ValueError(f"Invalid user field: {key}.")

        [result] = await (
            self._db()
            .update(AuthUser)
            .set(**updates)
            .where(AuthUser.c.id == str(user_id))
            .returning(AuthUser)
            .execute()
        )

        await self._invalidate_user_cache(user_id, user.email)
        if email is not None and email.lower() != user.email:
            await self._invalidate_user_cache(user_id, email)

        return self._to_user_info(result)

    async def delete_user(self, user_id: str | uuid.UUID) -> None:
        """Delete a user and all their sessions."""
        user = await self._fetch_user(user_id)
        if not user:
            raise UserNotFoundError()

        await self.sign_out_all(user_id)

        await self._db().delete(AuthUser).where(AuthUser.c.id == str(user_id)).execute()

        await self._invalidate_user_cache(user_id, user.email)

    async def count_users(self) -> int:
        """Return the total number of users."""
        return await self._db().select(AuthUser).count()

    # =========================================================================
    # Email/Password Authentication
    # =========================================================================

    async def sign_up(
        self,
        *,
        email: str,
        password: str,
        confirmation_url: str | None = None,
        confirmation_subject: str = "Confirm your email address",
        user_agent: str | None = None,
        ip_address: str | None = None,
        **kwargs: Any,
    ) -> tuple[UserInfo, TokenPair]:
        """Register a new user with email and password.

        Returns:
            Tuple of (user, token_pair)

        Raises:
            SignupDisabledError: If signup is disabled
            UserAlreadyExistsError: If user already exists
            PasswordValidationError: If password doesn't meet requirements
        """
        if not self._config.enable_signup:
            raise SignupDisabledError()
        if confirmation_url is None and self._config.enable_confirmation:
            raise ConfirmationURLMissingError(
                "`confirmation_url` is required when confirmation is enabled."
            )

        # Validate password
        validation = validate_password(self._config.password, password)
        if not validation.valid:
            raise PasswordValidationError("; ".join(validation.errors))

        # Check if user exists
        existing = await self._get_user_by_email(email=email.lower())
        if existing:
            raise UserAlreadyExistsError()

        # Create user
        hashed_password = await self._hasher.async_hash(password)
        now = datetime.now(UTC)

        email_confirmed_at = None if self._config.enable_confirmation else now

        vals: dict[str, Any] = {}
        for key, value in kwargs.items():
            if key in AuthUser.c:
                vals[key] = value
            else:
                raise ValueError(f"Invalid user field: {key}.")

        user = await (
            self._db()
            .insert(AuthUser)
            .values(
                email=email.lower(),
                encrypted_password=hashed_password,
                provider=AuthProvider.EMAIL,
                email_confirmed_at=email_confirmed_at,
                created_at=now,
                updated_at=now,
                last_sign_in_at=now,
                **vals,
            )
            .returning(AuthUser)
            .execute()
        )

        # Store confirmation token in KV and send email if needed
        if self._config.enable_confirmation:
            confirmation_token = generate_secure_token()
            ttl = self._config.confirmation_token_expire_hours * 3600
            await self._kv().set(
                f"{self._config.cache_prefix}:confirmation:{confirmation_token}".encode(),
                str(user.id).encode(),
                ttl=ttl,
            )

            await self._email().send_email(
                subject=confirmation_subject,
                to_email=email.lower(),
                template="confirmation.html",
                confirmation_url=f"{confirmation_url}?token={confirmation_token}",
            )

        # Create session and tokens
        token_pair = await self._create_session(
            user.id,
            role=user.role,
            user_agent=user_agent,
            ip_address=ip_address,
        )

        return self._to_user_info(user), token_pair

    async def sign_in_with_password(
        self,
        email: str,
        password: str,
        *,
        first_name: str | None = None,
        last_name: str | None = None,
        user_agent: str | None = None,
        ip_address: str | None = None,
    ) -> tuple[UserInfo, TokenPair]:
        """Sign in with email and password.

        Returns:
            Tuple of (user, token_pair)

        Raises:
            InvalidCredentialsError: If credentials are invalid
            UserNotActiveError: If user is disabled
            EmailNotConfirmedError: If email confirmation is required but not done
        """
        user = await self._get_user_by_email(email=email.lower())
        if not user:
            raise InvalidCredentialsError()

        if not user.encrypted_password:
            raise InvalidCredentialsError()

        if not await self._hasher.async_verify(password, user.encrypted_password):
            raise InvalidCredentialsError()

        if not user.is_active:
            raise UserNotActiveError()

        if self._config.enable_confirmation and not user.email_confirmed_at:
            raise EmailNotConfirmedError()

        # Update last sign in (and rehash password if needed) in a single write
        now = datetime.now(UTC)
        updates: dict[str, Any] = {"last_sign_in_at": now, "updated_at": now}

        if self._hasher.needs_rehash(user.encrypted_password):
            updates["encrypted_password"] = await self._hasher.async_hash(password)

        [user] = await (
            self._db()
            .update(AuthUser)
            .set(**updates)
            .where(AuthUser.c.id == user.id)
            .returning(AuthUser)
            .execute()
        )

        await self._invalidate_user_cache(user.id, user.email)

        token_pair = await self._create_session(
            user.id,
            role=user.role,
            user_agent=user_agent,
            ip_address=ip_address,
        )

        return self._to_user_info(user), token_pair

    # =========================================================================
    # Magic Link Authentication
    # =========================================================================

    async def sign_in_with_magic_link(self, *, email: str, magic_link_url: str) -> None:
        """Send a magic link email for passwordless sign in.

        Creates user if they don't exist (if signup enabled).

        Raises:
            ValueError: If magic link authentication is not enabled
        """
        if not self._config.enable_magic_link:
            raise ValueError("Magic link authentication is not enabled.")

        user = await self._get_user_by_email(email=email.lower())

        if not user:
            if not self._config.enable_signup:
                raise SignupDisabledError()

            # Create user for magic link
            now = datetime.now(UTC)
            user = await (
                self._db()
                .insert(AuthUser)
                .values(
                    email=email.lower(),
                    provider=AuthProvider.MAGIC_LINK,
                    email_confirmed_at=now,  # Magic link confirms email
                    created_at=now,
                    updated_at=now,
                )
                .returning(AuthUser)
                .execute()
            )

        # Store magic link token in KV (keyed by user ID)
        token = generate_secure_token()
        ttl = self._config.magic_link_expire_minutes * 60
        await self._kv().set(
            f"{self._config.cache_prefix}:magic_link:{token}".encode(),
            str(user.id).encode(),
            ttl=ttl,
        )

        # Send email
        await self._email().send_email(
            subject="Sign in to your account",
            to_email=email.lower(),
            template="magic_link.html",
            magic_link_url=f"{magic_link_url}?token={token}",
        )

    async def verify_magic_link(
        self,
        token: str,
        *,
        user_agent: str | None = None,
        ip_address: str | None = None,
    ) -> tuple[UserInfo, TokenPair]:
        """Verify a magic link and sign in.

        Returns:
            Tuple of (user, token_pair)

        Raises:
            MagicLinkExpiredError: If the magic link has expired or is invalid
            UserNotFoundError: If the user doesn't exist
            UserNotActiveError: If the user is disabled
        """
        kv_key = f"{self._config.cache_prefix}:magic_link:{token}".encode()
        user_id_bytes = await self._kv().get(kv_key)

        if user_id_bytes is None:
            raise MagicLinkExpiredError()

        # Delete on use (single use)
        await self._kv().delete(kv_key)

        user = await self._fetch_user(user_id_bytes.decode())
        if not user:
            raise UserNotFoundError()

        if not user.is_active:
            raise UserNotActiveError()

        # Update user
        now = datetime.now(UTC)
        updates: dict[str, Any] = {
            "last_sign_in_at": now,
            "updated_at": now,
        }
        # Magic link confirms email
        if not user.email_confirmed_at:
            updates["email_confirmed_at"] = now

        [user] = await (
            self._db()
            .update(AuthUser)
            .set(**updates)
            .where(AuthUser.c.id == user.id)
            .returning(AuthUser)
            .execute()
        )

        await self._invalidate_user_cache(user.id, user.email)

        token_pair = await self._create_session(
            user.id,
            role=user.role,
            user_agent=user_agent,
            ip_address=ip_address,
        )

        return self._to_user_info(user), token_pair

    # =========================================================================
    # OAuth Authentication
    # =========================================================================

    def get_oauth_provider(self, provider: str | AuthProvider) -> BaseOAuthProvider:
        if isinstance(provider, str):
            provider = AuthProvider(provider)
        oauth_provider = self._oauth_providers.get(provider)
        if oauth_provider is None:
            raise ValueError(f"OAuth provider not configured: {provider}")
        return oauth_provider

    def get_oauth_authorization_url(
        self,
        provider: str | AuthProvider,
        state: str,
        scopes: list[str] | None = None,
        redirect_uri: str | None = None,
    ) -> str:
        """Get the OAuth authorization URL for a provider.

        Args:
            provider_name: Name of the OAuth provider
            state: CSRF protection state token
            scopes: Optional scopes to request
            redirect_uri: Optional redirect URI override

        Returns:
            Authorization URL
        """
        oauth_provider = self.get_oauth_provider(provider)
        return oauth_provider.get_authorization_url(state, scopes, redirect_uri)

    async def sign_in_with_oauth(
        self,
        provider: str | AuthProvider,
        code: str,
        *,
        redirect_uri: str | None = None,
        user_agent: str | None = None,
        ip_address: str | None = None,
    ) -> tuple[UserInfo, TokenPair]:
        """Complete OAuth sign in with authorization code.

        Creates user if they don't exist.

        Returns:
            Tuple of (user, token_pair)

        Raises:
            ValueError: If provider is not registered
            OAuthError: If OAuth flow fails
            UserNotActiveError: If user is disabled
        """
        oauth_provider = self.get_oauth_provider(provider)

        # Get user info from provider
        user_info = await oauth_provider.authenticate(code, redirect_uri)

        # Find or create user
        user = await self._get_user_by_email(email=user_info.email)
        now = datetime.now(UTC)

        if user:
            # Update existing user
            if not user.is_active:
                raise UserNotActiveError()

            updates: dict[str, Any] = {
                "last_sign_in_at": now,
                "updated_at": now,
            }
            # Update provider info if first time with this provider
            if user.provider == AuthProvider.EMAIL:
                updates["provider"] = provider
                updates["provider_id"] = user_info.id

            # Confirm email if provider verified it
            if user_info.email_verified and not user.email_confirmed_at:
                updates["email_confirmed_at"] = now

            [user] = await (
                self._db()
                .update(AuthUser)
                .set(**updates)
                .where(AuthUser.c.id == user.id)
                .returning(AuthUser)
                .execute()
            )

            await self._invalidate_user_cache(user.id, user.email)
        else:
            user = await (
                self._db()
                .insert(AuthUser)
                .values(
                    email=user_info.email.lower(),
                    provider=provider,
                    provider_id=user_info.id,
                    email_confirmed_at=now if user_info.email_verified else None,
                    created_at=now,
                    updated_at=now,
                    last_sign_in_at=now,
                )
                .returning(AuthUser)
                .execute()
            )

        token_pair = await self._create_session(
            user.id,
            role=user.role,
            user_agent=user_agent,
            ip_address=ip_address,
        )

        return self._to_user_info(user), token_pair

    # =========================================================================
    # Session Management
    # =========================================================================

    async def _create_session(
        self,
        user_id: uuid.UUID,
        *,
        role: str = "default",
        user_agent: str | None = None,
        ip_address: str | None = None,
    ) -> TokenPair:
        """Create a new session and return tokens."""
        now = datetime.now(UTC)
        not_after = now + timedelta(days=self._config.session_expire_days)
        refresh_token = generate_secure_token()

        session = await (
            self._db()
            .insert(AuthSession)
            .values(
                user_id=user_id,
                token=refresh_token,
                role=role,
                user_agent=user_agent,
                ip_address=ip_address,
                not_after=not_after,
                created_at=now,
            )
            .returning(AuthSession)
            .execute()
        )

        return create_token_pair(
            self._config.jwt,
            user_id,
            session.session_id,
            refresh_token,
            extra_claims={"role": role},
        )

    async def refresh_token(self, refresh_token: str) -> TokenPair:
        """Refresh an access token using a refresh token.

        Implements token rotation for security. Happy path is 2 DB calls:
        one UPDATE…RETURNING to atomically revoke the old token, one INSERT
        for the new token.

        Returns:
            New TokenPair

        Raises:
            RefreshTokenRevokedError: If token was revoked
            RefreshTokenReusedError: If token reuse detected (potential theft)
            SessionExpiredError: If session has expired
        """
        # Atomically revoke and return the token in one query.
        # If the token doesn't exist or is already revoked, this returns [].
        revoked_rows = await (
            self._db()
            .update(AuthSession)
            .set(revoked=True)
            .eq(AuthSession.c.token, refresh_token)
            .not_(AuthSession.c.revoked)
            .returning(AuthSession)
            .execute()
        )

        if not revoked_rows:
            # Token not found or already revoked — check which case.
            existing = await (
                self._db()
                .select(AuthSession)
                .eq(AuthSession.c.token, refresh_token)
                .first_or_none()
            )
            if existing is not None and existing.revoked:
                # Reuse detected — revoke all tokens for this session.
                await (
                    self._db()
                    .update(AuthSession)
                    .set(revoked=True)
                    .eq(AuthSession.c.session_id, existing.session_id)
                    .execute()
                )
                if self._kv_client is not None:
                    await self._kv_client.delete(
                        f"{self._config.cache_prefix}:session:{existing.session_id}".encode()
                    )
                raise RefreshTokenReusedError()
            raise RefreshTokenRevokedError("Invalid refresh token")

        [token_record] = revoked_rows

        if token_record.not_after < datetime.now(UTC):
            raise SessionExpiredError()

        # Insert rotated token
        new_refresh_token = generate_secure_token()
        await (
            self._db()
            .insert(AuthSession)
            .values(
                user_id=token_record.user_id,
                session_id=token_record.session_id,
                token=new_refresh_token,
                role=token_record.role,
                user_agent=token_record.user_agent,
                ip_address=token_record.ip_address,
                not_after=token_record.not_after,
                created_at=datetime.now(UTC),
            )
            .execute()
        )

        # Invalidate stale session cache so next authenticate re-fetches
        if self._kv_client is not None:
            await self._kv_client.delete(
                f"{self._config.cache_prefix}:session:{token_record.session_id}".encode()
            )

        return create_token_pair(
            self._config.jwt,
            token_record.user_id,
            token_record.session_id,
            new_refresh_token,
            extra_claims={"role": token_record.role},
        )

    async def authenticate(self, request: AuthRequest) -> SessionInfo | None:
        """Authenticate a request via JWT (networkless).

        Extracts the Bearer token from the Authorization header,
        decodes and verifies the JWT signature and expiry.
        Returns ``SessionInfo`` built from JWT claims, or ``None``
        if the token is missing, invalid, or expired.
        """
        auth_header = request.headers.get("Authorization", "")
        if not auth_header.startswith("Bearer "):
            return None
        token = auth_header.removeprefix("Bearer ")
        payload = decode_token(self._config.jwt, token)
        if payload is None:
            return None
        extra = payload.extra or {}
        return SessionInfo(
            user_id=payload.sub,
            session_id=payload.session_id,
            role=extra.get("role", "default"),
            expires_at=payload.exp,
            metadata=extra,
            org_id=extra.get("org_id"),
            org_role=extra.get("org_role"),
        )

    async def list_sessions(
        self,
        *,
        user_id: str | uuid.UUID | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> list[AuthSession]:
        """List active (non-revoked) sessions ordered by creation date."""
        q = (
            self._db()
            .select(AuthSession)
            .where(~AuthSession.c.revoked)
            .order_by(AuthSession.c.created_at, asc=False)
        )
        if user_id is not None:
            q = q.where(AuthSession.c.user_id == str(user_id))
        if limit is not None:
            q = q.limit(limit)
        if offset is not None:
            q = q.offset(offset)
        return await q.execute()

    async def sign_out(self, session_id: str | uuid.UUID) -> None:
        """Sign out by deleting all tokens for a session."""
        await (
            self._db()
            .delete(AuthSession)
            .where(AuthSession.c.session_id == str(session_id))
            .execute()
        )

        # Invalidate session cache
        if self._kv_client is not None:
            await self._kv_client.delete(
                f"{self._config.cache_prefix}:session:{session_id}".encode()
            )

    async def sign_out_all(self, user_id: str | uuid.UUID) -> None:
        """Sign out all sessions for a user by deleting all tokens."""
        sessions = await (
            self._db()
            .select(AuthSession)
            .where(AuthSession.c.user_id == str(user_id))
            .execute()
        )

        if not sessions:
            return

        session_ids = {str(s.session_id) for s in sessions}

        await (
            self._db()
            .delete(AuthSession)
            .where(AuthSession.c.user_id == str(user_id))
            .execute()
        )

        # Invalidate all session caches
        if self._kv_client is not None:
            cache_keys = [
                f"{self._config.cache_prefix}:session:{sid}".encode()
                for sid in session_ids
            ]
            await self._kv_client.delete_many(cache_keys)

    # =========================================================================
    # Password Recovery
    # =========================================================================

    async def request_password_recovery(
        self,
        *,
        email: str,
        recovery_url: str,
        recovery_subject: str = "Reset your password",
        **kwargs: Any,
    ) -> None:
        """Send a password recovery email.

        Does not reveal whether user exists for security.
        """
        user = await self._get_user_by_email(email=email)
        if not user:
            return  # Don't reveal user doesn't exist

        if not user.is_active:
            return  # Don't reveal user is disabled

        # Store recovery token in KV
        token = generate_secure_token()
        ttl = self._config.recovery_token_expire_minutes * 60
        await self._kv().set(
            f"{self._config.cache_prefix}:recovery:{token}".encode(),
            str(user.id).encode(),
            ttl=ttl,
        )

        await self._email().send_email(
            subject=recovery_subject,
            to_email=email.lower(),
            template="recovery.html",
            recovery_url=f"{recovery_url}?token={token}",
            **kwargs,
        )

    async def reset_password(self, token: str, new_password: str) -> UserInfo:
        """Reset password using recovery token.

        Returns:
            Updated user

        Raises:
            RecoveryTokenInvalidError: If token is invalid or expired
            PasswordValidationError: If new password doesn't meet requirements
        """
        # Validate password
        validation = validate_password(self._config.password, new_password)
        if not validation.valid:
            raise PasswordValidationError("; ".join(validation.errors))

        # Look up recovery token in KV
        kv_key = f"{self._config.cache_prefix}:recovery:{token}".encode()
        user_id_bytes = await self._kv().get(kv_key)

        if user_id_bytes is None:
            raise RecoveryTokenInvalidError()

        # Delete token (single use)
        await self._kv().delete(kv_key)

        user = await self._fetch_user(user_id=user_id_bytes.decode())
        if user is None:
            raise RecoveryTokenInvalidError()

        # Update password
        hashed_password = await self._hasher.async_hash(new_password)
        now = datetime.now(UTC)

        [result] = await (
            self._db()
            .update(AuthUser)
            .set(encrypted_password=hashed_password, updated_at=now)
            .where(AuthUser.c.id == user.id)
            .returning(AuthUser)
            .execute()
        )

        await self._invalidate_user_cache(user.id, user.email)

        # Sign out all sessions (security measure)
        await self.sign_out_all(user.id)

        return self._to_user_info(result)

    # =========================================================================
    # Email Confirmation
    # =========================================================================

    async def confirm_email(self, token: str) -> UserInfo:
        """Confirm email address with token.

        Returns:
            Updated user

        Raises:
            ConfirmationTokenInvalidError: If token is invalid or expired
        """
        kv_key = f"{self._config.cache_prefix}:confirmation:{token}".encode()
        user_id_bytes = await self._kv().get(kv_key)

        if user_id_bytes is None:
            raise ConfirmationTokenInvalidError()

        # Delete token (single use)
        await self._kv().delete(kv_key)

        user = await self._fetch_user(user_id=user_id_bytes.decode())
        if user is None:
            raise ConfirmationTokenInvalidError()

        # Confirm email
        now = datetime.now(UTC)
        [result] = await (
            self._db()
            .update(AuthUser)
            .set(email_confirmed_at=now, updated_at=now)
            .where(AuthUser.c.id == user.id)
            .returning(AuthUser)
            .execute()
        )

        await self._invalidate_user_cache(user.id, user.email)

        return self._to_user_info(result)

    async def resend_confirmation_email(
        self,
        *,
        email: str,
        confirmation_url: str,
        confirmation_subject: str = "Confirm your email address",
        **kwargs: Any,
    ) -> None:
        """Resend email confirmation.

        Does not reveal whether user exists for security.
        """
        user = await self._get_user_by_email(email=email)
        if not user:
            return

        if user.email_confirmed_at:
            return  # Already confirmed

        # Store new confirmation token in KV
        token = generate_secure_token()
        ttl = self._config.confirmation_token_expire_hours * 3600
        await self._kv().set(
            f"{self._config.cache_prefix}:confirmation:{token}".encode(),
            str(user.id).encode(),
            ttl=ttl,
        )

        await self._email().send_email(
            subject=confirmation_subject,
            to_email=email.lower(),
            template="confirmation.html",
            confirmation_url=f"{confirmation_url}?token={token}",
            **kwargs,
        )

    # =========================================================================
    # Organizations
    # =========================================================================

    def _to_org_info(self, org: AuthOrganization) -> OrgInfo:
        """Convert an AuthOrganization ORM model to a public OrgInfo."""
        import json

        metadata: dict[str, Any] = {}
        if org.metadata:
            metadata = json.loads(org.metadata)
        return OrgInfo(
            id=str(org.id),
            name=org.name,
            slug=org.slug,
            metadata=metadata,
            created_at=org.created_at,
            updated_at=org.updated_at,
        )

    def _to_org_member_info(self, member: AuthOrgMember) -> OrgMemberInfo:
        """Convert an AuthOrgMember ORM model to a public OrgMemberInfo."""
        return OrgMemberInfo(
            org_id=str(member.org_id),
            user_id=str(member.user_id),
            role=member.role,
            created_at=member.created_at,
            updated_at=member.updated_at,
        )

    async def create_org(
        self,
        *,
        name: str,
        slug: str,
        creator_id: str | uuid.UUID,
        **kwargs: Any,
    ) -> OrgInfo:
        """Create an organization. The creator is added as owner."""
        # Check for duplicate slug
        existing = await (
            self._db()
            .select(AuthOrganization)
            .where(AuthOrganization.c.slug == slug)
            .first_or_none()
        )
        if existing is not None:
            raise OrgAlreadyExistsError()

        now = datetime.now(UTC)
        org = await (
            self._db()
            .insert(AuthOrganization)
            .values(
                name=name,
                slug=slug,
                created_at=now,
                updated_at=now,
            )
            .returning(AuthOrganization)
            .execute()
        )

        # Add creator as owner
        await (
            self._db()
            .insert(AuthOrgMember)
            .values(
                org_id=org.id,
                user_id=str(creator_id),
                role="owner",
                created_at=now,
                updated_at=now,
            )
            .execute()
        )

        return self._to_org_info(org)

    async def get_org(self, org_id: str | uuid.UUID) -> OrgInfo | None:
        """Get an organization by ID."""
        org = await (
            self._db()
            .select(AuthOrganization)
            .where(AuthOrganization.c.id == str(org_id))
            .first_or_none()
        )
        return self._to_org_info(org) if org is not None else None

    async def get_org_by_slug(self, slug: str) -> OrgInfo | None:
        """Get an organization by slug."""
        org = await (
            self._db()
            .select(AuthOrganization)
            .where(AuthOrganization.c.slug == slug)
            .first_or_none()
        )
        return self._to_org_info(org) if org is not None else None

    async def update_org(
        self,
        *,
        org_id: str | uuid.UUID,
        name: str | None = None,
        slug: str | None = None,
        **kwargs: Any,
    ) -> OrgInfo:
        """Update an organization."""
        existing = await (
            self._db()
            .select(AuthOrganization)
            .where(AuthOrganization.c.id == str(org_id))
            .first_or_none()
        )
        if existing is None:
            raise OrgNotFoundError()

        updates: dict[str, Any] = {"updated_at": datetime.now(UTC)}
        if name is not None:
            updates["name"] = name
        if slug is not None:
            updates["slug"] = slug

        [result] = await (
            self._db()
            .update(AuthOrganization)
            .set(**updates)
            .where(AuthOrganization.c.id == str(org_id))
            .returning(AuthOrganization)
            .execute()
        )

        return self._to_org_info(result)

    async def delete_org(self, org_id: str | uuid.UUID) -> None:
        """Delete an organization and all its memberships."""
        existing = await (
            self._db()
            .select(AuthOrganization)
            .where(AuthOrganization.c.id == str(org_id))
            .first_or_none()
        )
        if existing is None:
            raise OrgNotFoundError()

        await (
            self._db()
            .delete(AuthOrganization)
            .where(AuthOrganization.c.id == str(org_id))
            .execute()
        )

    async def list_orgs(
        self,
        *,
        user_id: str | uuid.UUID | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> list[OrgInfo]:
        """List organizations, optionally filtered by user membership."""
        if user_id is not None:
            # Get org IDs the user belongs to, then fetch orgs
            members = await (
                self._db()
                .select(AuthOrgMember)
                .where(AuthOrgMember.c.user_id == str(user_id))
                .execute()
            )
            org_ids = [str(m.org_id) for m in members]
            if not org_ids:
                return []
            q = (
                self._db()
                .select(AuthOrganization)
                .where(AuthOrganization.c.id.in_(org_ids))
                .order_by(AuthOrganization.c.created_at, asc=False)
            )
        else:
            q = (
                self._db()
                .select(AuthOrganization)
                .order_by(AuthOrganization.c.created_at, asc=False)
            )

        if limit is not None:
            q = q.limit(limit)
        if offset is not None:
            q = q.offset(offset)
        return [self._to_org_info(o) for o in await q.execute()]

    # =========================================================================
    # Organization Membership
    # =========================================================================

    async def add_org_member(
        self,
        *,
        org_id: str | uuid.UUID,
        user_id: str | uuid.UUID,
        role: str = "member",
    ) -> OrgMemberInfo:
        """Add a user to an organization."""
        existing = await (
            self._db()
            .select(AuthOrgMember)
            .where(AuthOrgMember.c.org_id == str(org_id))
            .where(AuthOrgMember.c.user_id == str(user_id))
            .first_or_none()
        )
        if existing is not None:
            raise OrgMemberExistsError()

        now = datetime.now(UTC)
        member = await (
            self._db()
            .insert(AuthOrgMember)
            .values(
                org_id=str(org_id),
                user_id=str(user_id),
                role=role,
                created_at=now,
                updated_at=now,
            )
            .returning(AuthOrgMember)
            .execute()
        )

        return self._to_org_member_info(member)

    async def update_org_member(
        self,
        *,
        org_id: str | uuid.UUID,
        user_id: str | uuid.UUID,
        role: str,
    ) -> OrgMemberInfo:
        """Update a member's role within an organization."""
        existing = await (
            self._db()
            .select(AuthOrgMember)
            .where(AuthOrgMember.c.org_id == str(org_id))
            .where(AuthOrgMember.c.user_id == str(user_id))
            .first_or_none()
        )
        if existing is None:
            raise OrgMemberNotFoundError()

        [result] = await (
            self._db()
            .update(AuthOrgMember)
            .set(role=role, updated_at=datetime.now(UTC))
            .where(AuthOrgMember.c.org_id == str(org_id))
            .where(AuthOrgMember.c.user_id == str(user_id))
            .returning(AuthOrgMember)
            .execute()
        )

        return self._to_org_member_info(result)

    async def remove_org_member(
        self,
        *,
        org_id: str | uuid.UUID,
        user_id: str | uuid.UUID,
    ) -> None:
        """Remove a user from an organization."""
        existing = await (
            self._db()
            .select(AuthOrgMember)
            .where(AuthOrgMember.c.org_id == str(org_id))
            .where(AuthOrgMember.c.user_id == str(user_id))
            .first_or_none()
        )
        if existing is None:
            raise OrgMemberNotFoundError()

        # Prevent removing the last owner
        if existing.role == "owner":
            owner_count = await (
                self._db()
                .select(AuthOrgMember)
                .where(AuthOrgMember.c.org_id == str(org_id))
                .where(AuthOrgMember.c.role == "owner")
                .count()
            )
            if owner_count <= 1:
                raise OrgLastOwnerError()

        await (
            self._db()
            .delete(AuthOrgMember)
            .where(AuthOrgMember.c.org_id == str(org_id))
            .where(AuthOrgMember.c.user_id == str(user_id))
            .execute()
        )

    async def list_org_members(
        self,
        org_id: str | uuid.UUID,
        *,
        limit: int | None = None,
        offset: int | None = None,
    ) -> list[OrgMemberInfo]:
        """List members of an organization."""
        q = (
            self._db()
            .select(AuthOrgMember)
            .where(AuthOrgMember.c.org_id == str(org_id))
            .order_by(AuthOrgMember.c.created_at, asc=True)
        )
        if limit is not None:
            q = q.limit(limit)
        if offset is not None:
            q = q.offset(offset)
        return [self._to_org_member_info(m) for m in await q.execute()]

    async def get_org_member(
        self,
        *,
        org_id: str | uuid.UUID,
        user_id: str | uuid.UUID,
    ) -> OrgMemberInfo | None:
        """Get a single membership record."""
        member = await (
            self._db()
            .select(AuthOrgMember)
            .where(AuthOrgMember.c.org_id == str(org_id))
            .where(AuthOrgMember.c.user_id == str(user_id))
            .first_or_none()
        )
        return self._to_org_member_info(member) if member is not None else None

    # =========================================================================
    # Organization Session Context
    # =========================================================================

    async def set_active_org(
        self,
        *,
        session_id: str | uuid.UUID,
        org_id: str | uuid.UUID | None,
    ) -> TokenPair:
        """Switch the active organization for a session. Returns new tokens."""
        # Find the active session
        session = await (
            self._db()
            .select(AuthSession)
            .where(AuthSession.c.session_id == str(session_id))
            .where(~AuthSession.c.revoked)
            .order_by(AuthSession.c.created_at, asc=False)
            .first_or_none()
        )
        if session is None:
            raise SessionNotFoundError()

        extra_claims: dict[str, Any] = {"role": session.role}

        if org_id is not None:
            # Verify user is a member
            member = await (
                self._db()
                .select(AuthOrgMember)
                .where(AuthOrgMember.c.org_id == str(org_id))
                .where(AuthOrgMember.c.user_id == str(session.user_id))
                .first_or_none()
            )
            if member is None:
                raise NotOrgMemberError()

            extra_claims["org_id"] = str(org_id)
            extra_claims["org_role"] = member.role

            # Update session's org_id
            await (
                self._db()
                .update(AuthSession)
                .set(org_id=str(org_id))
                .where(AuthSession.c.session_id == str(session_id))
                .where(~AuthSession.c.revoked)
                .execute()
            )
        else:
            # Clear org context
            await (
                self._db()
                .update(AuthSession)
                .set(org_id=None)
                .where(AuthSession.c.session_id == str(session_id))
                .where(~AuthSession.c.revoked)
                .execute()
            )

        return create_token_pair(
            self._config.jwt,
            session.user_id,
            session.session_id,
            session.token,
            extra_claims=extra_claims,
        )



"""Check command - validate schema is synced with snapshots."""

from __future__ import annotations

from pathlib import Path

import typer

from derp.config import ConfigError, DerpConfig
from derp.orm.loader import discover_tables
from derp.orm.migrations.journal import load_journal, load_latest_snapshot
from derp.orm.migrations.snapshot.differ import SnapshotDiffer
from derp.orm.migrations.snapshot.models import SchemaSnapshot
from derp.orm.migrations.snapshot.serializer import serialize_schema
from derp.orm.migrations.statements.types import (
    AddColumnStatement,
    AlterColumnDefaultStatement,
    AlterColumnNullableStatement,
    AlterColumnTypeStatement,
    AlterEnumAddValueStatement,
    CreateEnumStatement,
    CreateForeignKeyStatement,
    CreateIndexStatement,
    CreatePolicyStatement,
    CreateSchemaStatement,
    CreateSequenceStatement,
    CreateTableStatement,
    CreateUniqueConstraintStatement,
    DisableRLSStatement,
    DropColumnStatement,
    DropEnumStatement,
    DropForeignKeyStatement,
    DropIndexStatement,
    DropPolicyStatement,
    DropSchemaStatement,
    DropSequenceStatement,
    DropTableStatement,
    DropUniqueConstraintStatement,
    EnableRLSStatement,
    RenameColumnStatement,
    RenameTableStatement,
)


def check() -> None:
    """Check for schema changes without generating migration.

    Compares your current schema definition against the latest snapshot
    and reports any differences. Useful for CI/CD pipelines to verify
    schema is in sync before deployment.

    Exit codes:
      0 - Schema is up to date
      1 - Schema changes detected (migration needed)
    """
    try:
        config = DerpConfig.load()
    except ConfigError as exc:
        typer.echo(f"Error: {exc}", err=True)
        raise typer.Exit(1)

    migrations_dir = Path(config.database.migrations_dir)
    schema_path = config.database.schema_path

    # Load tables from schema module
    try:
        tables = discover_tables(
            schema_path,
            include_auth=config.auth is not None and config.auth.native is not None,
        )
    except FileNotFoundError:
        typer.echo(f"Error: Schema file not found: {schema_path}", err=True)
        raise typer.Exit(1)

    if not tables:
        typer.echo(f"No Table classes found in {schema_path}", err=True)
        raise typer.Exit(1)

    # Load journal and previous snapshot
    journal = load_journal(migrations_dir)
    prev_snapshot_data = load_latest_snapshot(migrations_dir, journal)

    if prev_snapshot_data is None:
        typer.echo(
            "No snapshots found. Run 'derp generate' to create initial migration.",
            err=True,
        )
        raise typer.Exit(1)

    prev_snapshot = SchemaSnapshot.model_validate(prev_snapshot_data)

    # Serialize current schema
    current_snapshot = serialize_schema(tables, schema="public")

    # Diff
    differ = SnapshotDiffer(prev_snapshot, current_snapshot)
    statements = differ.diff()

    if not statements:
        typer.echo("Schema is up to date. No changes detected.")
        raise typer.Exit(0)

    # Report changes
    typer.echo("Schema changes detected:", err=True)
    typer.echo("", err=True)

    for stmt in statements:
        description = _describe_statement(stmt)
        typer.echo(f"  - {description}", err=True)

    typer.echo("", err=True)
    typer.echo(f"Total: {len(statements)} change(s)", err=True)
    typer.echo("", err=True)
    typer.echo("Run 'derp generate' to create a migration.", err=True)

    raise typer.Exit(1)


def _describe_statement(stmt) -> str:
    """Generate human-readable description of a statement."""
    match stmt:
        case CreateTableStatement():
            return f"CREATE TABLE {stmt.table_name}"
        case DropTableStatement():
            return f"DROP TABLE {stmt.table_name}"
        case RenameTableStatement():
            return f"RENAME TABLE {stmt.from_table} -> {stmt.to_table}"
        case AddColumnStatement():
            return f"ADD COLUMN {stmt.table_name}.{stmt.column.name}"
        case DropColumnStatement():
            return f"DROP COLUMN {stmt.table_name}.{stmt.column_name}"
        case RenameColumnStatement():
            return (
                f"RENAME COLUMN {stmt.table_name}.{stmt.from_column} "
                f"-> {stmt.to_column}"
            )
        case AlterColumnTypeStatement():
            return (
                f"ALTER COLUMN {stmt.table_name}.{stmt.column_name} "
                f"TYPE {stmt.new_type}"
            )
        case AlterColumnNullableStatement():
            nullable = "NULL" if stmt.nullable else "NOT NULL"
            return f"ALTER COLUMN {stmt.table_name}.{stmt.column_name} SET {nullable}"
        case AlterColumnDefaultStatement():
            if stmt.default is None:
                return f"ALTER COLUMN {stmt.table_name}.{stmt.column_name} DROP DEFAULT"
            return f"ALTER COLUMN {stmt.table_name}.{stmt.column_name} SET DEFAULT"
        case CreateForeignKeyStatement():
            return f"ADD FOREIGN KEY {stmt.name}"
        case DropForeignKeyStatement():
            return f"DROP FOREIGN KEY {stmt.name}"
        case CreateUniqueConstraintStatement():
            return f"ADD UNIQUE CONSTRAINT {stmt.name}"
        case DropUniqueConstraintStatement():
            return f"DROP UNIQUE CONSTRAINT {stmt.name}"
        case CreateIndexStatement():
            return f"CREATE INDEX {stmt.name}"
        case DropIndexStatement():
            return f"DROP INDEX {stmt.name}"
        case CreateEnumStatement():
            return f"CREATE ENUM {stmt.name}"
        case DropEnumStatement():
            return f"DROP ENUM {stmt.name}"
        case AlterEnumAddValueStatement():
            return f"ALTER ENUM {stmt.name} ADD VALUE '{stmt.value}'"
        case CreateSequenceStatement():
            return f"CREATE SEQUENCE {stmt.name}"
        case DropSequenceStatement():
            return f"DROP SEQUENCE {stmt.name}"
        case CreateSchemaStatement():
            return f"CREATE SCHEMA {stmt.name}"
        case DropSchemaStatement():
            return f"DROP SCHEMA {stmt.name}"
        case EnableRLSStatement():
            return f"ENABLE RLS ON {stmt.table_name}"
        case DisableRLSStatement():
            return f"DISABLE RLS ON {stmt.table_name}"
        case CreatePolicyStatement():
            return f"CREATE POLICY {stmt.name} ON {stmt.table_name}"
        case DropPolicyStatement():
            return f"DROP POLICY {stmt.name} ON {stmt.table_name}"
        case _:
            return f"{stmt.type}"



"""Derp client for interacting with database, file storage, and more."""

from __future__ import annotations

import datetime
from types import TracebackType
from typing import Self

from derp.auth.base import BaseAuthClient
from derp.auth.clerk_client import ClerkAuthClient
from derp.auth.email import EmailClient
from derp.auth.native_client import NativeAuthClient
from derp.config import DerpConfig
from derp.kv.base import KVClient
from derp.kv.valkey import ValkeyClient
from derp.orm import DatabaseEngine
from derp.orm.router import ReplicaRouter
from derp.payments import PaymentsClient
from derp.queue.base import QueueClient, Schedule, ScheduleType
from derp.queue.celery import CeleryQueueClient
from derp.queue.vercel import VercelQueueClient
from derp.storage import StorageClient


class DerpClient:
    """Derp client for interacting with database, file storage, and more."""

    def __init__(self, config: DerpConfig):
        self._config: DerpConfig = config
        self._db: DatabaseEngine = DatabaseEngine(
            config.database.db_url,
            min_size=config.database.pool_min_size,
            max_size=config.database.pool_max_size,
            statement_cache_size=config.database.statement_cache_size,
        )
        self._replica_db: DatabaseEngine | None = (
            DatabaseEngine(
                config.database.replica_url,
                min_size=(
                    config.database.replica_pool_min_size
                    or config.database.pool_min_size
                ),
                max_size=(
                    config.database.replica_pool_max_size
                    or config.database.pool_max_size
                ),
                statement_cache_size=config.database.replica_statement_cache_size,
            )
            if config.database.replica_url is not None
            else None
        )
        self._email: EmailClient | None = (
            EmailClient(self._config.email) if self._config.email is not None else None
        )
        self._storage: StorageClient | None = (
            StorageClient(self._config.storage)
            if self._config.storage is not None
            else None
        )
        self._auth: BaseAuthClient | None = None
        if self._config.auth is not None:
            if self._config.auth.native is not None:
                self._auth = NativeAuthClient(self._config.auth.native)
            elif self._config.auth.clerk is not None:
                self._auth = ClerkAuthClient(self._config.auth.clerk)
        self._kv: KVClient | None = (
            ValkeyClient(self._config.kv.valkey)
            if self._config.kv is not None and self._config.kv.valkey is not None
            else None
        )
        self._payments: PaymentsClient | None = (
            PaymentsClient(self._config.payments)
            if self._config.payments is not None
            else None
        )
        self._queue: QueueClient | None = None
        if self._config.queue is not None:
            if self._config.queue.celery is not None:
                self._queue = CeleryQueueClient(self._config.queue.celery)
            elif self._config.queue.vercel is not None:
                self._queue = VercelQueueClient(self._config.queue.vercel)
            if self._queue is not None and self._config.queue.schedules:
                self._queue.register_schedules(
                    [
                        Schedule(
                            name=sc.name,
                            task=sc.task,
                            type=(
                                ScheduleType.CRON if sc.cron else ScheduleType.INTERVAL
                            ),
                            cron=sc.cron,
                            interval=(
                                datetime.timedelta(seconds=sc.interval_seconds)
                                if sc.interval_seconds
                                else None
                            ),
                            payload=sc.payload,
                            queue=sc.queue,
                            path=sc.path,
                        )
                        for sc in self._config.queue.schedules
                    ]
                )
        self._router: ReplicaRouter | None = None
        self._in_session = False

        if (
            self._config.auth is not None
            and self._config.auth.native is not None
            and self._email is None
        ):
            raise ValueError(
                "The email client needs to be configured for native authentication "
                "to work. Please make sure to configure `EmailConfig` when "
                "`NativeAuthConfig` is configured via `derp.toml` or `DerpConfig`."
            )

    async def connect(self) -> None:
        """Start a session."""
        await self._db.connect()
        if self._replica_db is not None:
            await self._replica_db.connect()
        if self._storage is not None:
            await self._storage.connect()
        if self._kv is not None:
            await self._kv.connect()
        if self._payments is not None:
            await self._payments.connect()
        if self._queue is not None:
            await self._queue.connect()
        if self._auth is not None:
            self._auth.set_db(self._db)
            self._auth.set_email(self._email)
            if self._kv is not None:
                self._auth.set_kv(self._kv)
        if self._kv is not None:
            self._db.set_cache(self._kv)

        # Set up replica router if replica is configured
        if self._replica_db is not None:
            self._router = ReplicaRouter(
                self._db.pool,
                self._replica_db.pool,
                self._config.database,
            )
            await self._router.start()
            self._db.set_router(self._router)

        self._in_session = True

    async def disconnect(
        self,
    ) -> None:
        """End a session."""
        errors: list[Exception] = []

        # Stop router before closing pools
        if self._router is not None:
            try:
                await self._router.stop()
            except Exception as exc:
                errors.append(exc)
            self._db.set_router(None)
            self._router = None

        for client in [
            self._db,
            self._replica_db,
            self._storage,
            self._kv,
            self._payments,
            self._queue,
        ]:
            if client is not None:
                try:
                    await client.disconnect()
                except Exception as exc:
                    errors.append(exc)

        if self._auth is not None:
            self._auth.set_db(None)
            self._auth.set_email(None)
            self._auth.set_kv(None)
        self._db.set_cache(None)

        self._in_session = False

        if errors:
            raise ExceptionGroup("errors during disconnect", errors)

    async def __aenter__(self) -> Self:
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.disconnect()

    @property
    def db(self) -> DatabaseEngine:
        """Get the database engine."""
        if not self._in_session:
            raise ValueError("Not in a session. Call `connect()` first.")
        return self._db

    @property
    def email(self) -> EmailClient:
        """Get the email client."""
        if not self._in_session:
            raise ValueError("Not in a session. Call `connect()` first.")
        if self._email is None:
            raise ValueError("`EmailConfig` was not passed to `DerpConfig`.")
        return self._email

    @property
    def storage(self) -> StorageClient:
        """Get the storage client."""
        if not self._in_session:
            raise ValueError("Not in a session. Call `connect()` first.")
        if self._storage is None:
            raise ValueError("`StorageConfig` was not passed to `DerpConfig`.")
        return self._storage

    @property
    def auth(self) -> BaseAuthClient:
        """Get the auth service."""
        if not self._in_session:
            raise ValueError("Not in a session. Call `connect()` first.")
        if self._auth is None:
            raise ValueError("`AuthConfig` was not passed to `DerpConfig`.")
        return self._auth

    @property
    def kv(self) -> KVClient:
        """Get the KV client."""
        if not self._in_session:
            raise ValueError("Not in a session. Call `connect()` first.")
        if self._kv is None:
            raise ValueError("`KVConfig` was not passed to `DerpConfig`.")
        return self._kv

    @property
    def payments(self) -> PaymentsClient:
        """Get the payments client."""
        if not self._in_session:
            raise ValueError("Not in a session. Call `connect()` first.")
        if self._payments is None:
            raise ValueError("`PaymentsConfig` was not passed to `DerpConfig`.")
        return self._payments

    @property
    def queue(self) -> QueueClient:
        """Get the queue client."""
        if not self._in_session:
            raise ValueError("Not in a session. Call `connect()` first.")
        if self._queue is None:
            raise ValueError("`QueueConfig` was not passed to `DerpConfig`.")
        return self._queue

    @property
    def config(self) -> DerpConfig:
        """Get the Derp configuration."""
        return self._config



"""Async query engine (asyncpg wrapper) for Derp ORM."""

from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from types import TracebackType
from typing import Any, overload

import asyncpg

from derp.kv.base import KVClient
from derp.orm.fields import FieldInfo
from derp.orm.query.builder import DeleteQuery, InsertQuery, SelectQuery, UpdateQuery
from derp.orm.query.expressions import Expression
from derp.orm.query.table_ref import TableRef
from derp.orm.router import ReplicaRouter
from derp.orm.table import Table


class Transaction:
    """Transaction context manager with query builder support.

    Queries created via this transaction's ``select``, ``insert``,
    ``update``, and ``delete`` methods reuse the transaction's
    connection instead of acquiring a new one from the pool.

    Example::

        async with db.transaction() as txn:
            user = await txn.insert(User).values(name="Alice").returning(User).execute()
            await txn.update(Profile).set(user_id=user.id).execute()
    """

    def __init__(self, connection: asyncpg.Connection):
        self._connection = connection
        self._transaction: asyncpg.connection.transaction.Transaction | None = None

    async def __aenter__(self) -> Transaction:
        self._transaction = self._connection.transaction()
        await self._transaction.start()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if self._transaction is None:
            return
        if exc_type is not None:
            await self._transaction.rollback()
        else:
            await self._transaction.commit()

    # Single table selection - returns model instances
    @overload
    def select[T: Table](self, table: type[T], /) -> SelectQuery[T]: ...

    # Fallback for column selections and mixed cases - returns dicts
    @overload
    def select(
        self, *columns: type[Table] | FieldInfo[Any] | Expression
    ) -> SelectQuery[dict[str, Any]]: ...

    def select(
        self, *columns: type[Table] | FieldInfo[Any] | Expression
    ) -> SelectQuery[Any]:
        """Start a SELECT query bound to this transaction's connection."""
        return SelectQuery(self._connection, columns)

    def insert[T: Table](self, table: type[T]) -> InsertQuery[T]:
        """Start an INSERT query bound to this transaction's connection."""
        return InsertQuery(self._connection, table)

    def update[T: Table](self, table: type[T]) -> UpdateQuery[T]:
        """Start an UPDATE query bound to this transaction's connection."""
        return UpdateQuery(self._connection, table)

    def delete[T: Table](self, table: type[T]) -> DeleteQuery[T]:
        """Start a DELETE query bound to this transaction's connection."""
        return DeleteQuery(self._connection, table)

    def table(self, table: str) -> TableRef:
        """Start a non ORM query from a string table name."""
        return TableRef(table, self._connection)


class DatabaseEngine:
    """Main async database engine for Derp ORM.

    Example:
        db = DatabaseEngine("postgresql://user:pass@localhost:5432/mydb")

        async with db:
            users = await db.select(User).where(User.name == "Alice").execute()

        # Or manual lifecycle
        await db.connect()
        users = await db.select(User).execute()
        await db.disconnect()
    """

    def __init__(
        self,
        dsn: str,
        *,
        min_size: int = 2,
        max_size: int = 10,
        statement_cache_size: int | None = None,
    ):
        """Initialize Derp engine.

        Args:
            dsn: PostgreSQL connection string
            min_size: Minimum connection pool size
            max_size: Maximum connection pool size
            statement_cache_size: Size of the prepared statement cache per
                connection. Set to 0 to disable, which is required when
                connecting through PgBouncer in transaction mode. None
                uses asyncpg's default.
        """
        self._dsn = dsn
        self._min_size = min_size
        self._max_size = max_size
        self._statement_cache_size = statement_cache_size
        self._pool: asyncpg.Pool | None = None
        self._cache_store: KVClient | None = None
        self._router: ReplicaRouter | None = None

    async def connect(self) -> None:
        """Establish connection pool."""
        if self._pool is not None:
            return
        kwargs: dict[str, Any] = {
            "min_size": self._min_size,
            "max_size": self._max_size,
        }
        if self._statement_cache_size is not None:
            kwargs["statement_cache_size"] = self._statement_cache_size
        self._pool = await asyncpg.create_pool(self._dsn, **kwargs)

    async def disconnect(self) -> None:
        """Close connection pool."""
        if self._pool is not None:
            await self._pool.close()
            self._pool = None

    async def __aenter__(self) -> DatabaseEngine:
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.disconnect()

    def set_cache(self, store: KVClient | None) -> None:
        """Set the KV store for query result caching."""
        self._cache_store = store

    def set_router(self, router: ReplicaRouter | None) -> None:
        """Set the replica router for automatic read routing."""
        self._router = router

    @property
    def pool(self) -> asyncpg.Pool:
        """Get the connection pool."""
        if self._pool is None:
            raise RuntimeError(
                "Database not connected. Call connect() or use async context manager."
            )
        return self._pool

    # Single table selection - returns model instances
    @overload
    def select[T: Table](self, table: type[T], /) -> SelectQuery[T]: ...

    # Fallback for column selections and mixed cases - returns dicts
    @overload
    def select(
        self, *columns: type[Table] | FieldInfo[Any] | Expression
    ) -> SelectQuery[dict[str, Any]]: ...

    def select(
        self, *columns: type[Table] | FieldInfo[Any] | Expression
    ) -> SelectQuery[Any]:
        """Start a SELECT query.

        Args:
            *columns: Table classes, FieldInfo columns, or Expressions to select

        Returns:
            Typed SelectQuery builder:
            - SelectQuery[T] when selecting a single Table class (returns list[T])
            - SelectQuery[dict[str, Any]] for column selections (returns list[dict])

        Examples:
            # Select all columns from User - returns list[User]
            db.select(User).where(User.c.id == 1)

            # Select specific columns - returns list[dict[str, Any]]
            db.select(User.c.id, User.c.name)

            # Join queries
            db.select(Post, User.c.name)
              .from_(Post)
              .inner_join(User, Post.c.author_id == User.c.id)

            # Non ORM queries
            db.table("users").select("*").eq("id", 1).execute()
        """
        return SelectQuery(
            self._pool, columns, cache_store=self._cache_store, router=self._router
        )

    def insert[T: Table](self, table: type[T]) -> InsertQuery[T]:
        """Start an INSERT query.

        Args:
            table: Table class to insert into

        Returns:
            InsertQuery builder

        Example::

            await (
                db.insert(User)
                .values(name="Bob", email="bob@example.com")
                .returning(User)
                .execute()
            )
        """
        return InsertQuery(self._pool, table, router=self._router)

    def update[T: Table](self, table: type[T]) -> UpdateQuery[T]:
        """Start an UPDATE query.

        Args:
            table: Table class to update

        Returns:
            UpdateQuery builder

        Example:
            await db.update(User).set(name="Robert").where(User.id == 1).execute()
        """
        return UpdateQuery(self._pool, table, router=self._router)

    def delete[T: Table](self, table: type[T]) -> DeleteQuery[T]:
        """Start a DELETE query.

        Args:
            table: Table class to delete from

        Returns:
            DeleteQuery builder

        Example:
            await db.delete(User).where(User.id == 1).execute()
        """
        return DeleteQuery(self._pool, table, router=self._router)

    def table(self, table: Table | str) -> TableRef:
        """Start a non ORM query from a string table name.

        Args:
            table: SQL table name

        Returns:
            TableRef with select/insert/update/delete methods

        Example:
            await db.table("users").select("*").eq("id", 1).execute()
        """
        return TableRef(
            table if isinstance(table, str) else table.get_table_name(),
            self._pool,
            cache_store=self._cache_store,
            router=self._router,
        )

    async def execute(
        self, query: str, params: list[Any] | None = None
    ) -> list[dict[str, Any]]:
        """Execute a raw SQL query.

        Args:
            query: SQL query string with $N placeholders
            params: Parameter values

        Returns:
            List of row dicts

        Example:
            result = await db.execute("SELECT * FROM users WHERE id = $1", [user_id])
        """
        async with self.pool.acquire() as conn:
            if params:
                rows = await conn.fetch(query, *params)
            else:
                rows = await conn.fetch(query)
            return [dict(row) for row in rows]

    async def execute_many(self, query: str, params_list: list[list[Any]]) -> None:
        """Execute a query with multiple parameter sets.

        Args:
            query: SQL query string
            params_list: List of parameter lists
        """
        async with self.pool.acquire() as conn:
            await conn.executemany(query, params_list)

    @asynccontextmanager
    async def transaction(self) -> AsyncIterator[Transaction]:
        """Start a transaction.

        Example:
            async with db.transaction():
                await db.insert(User).values(...).execute()
                await db.update(Post).set(...).execute()
        """
        async with self.pool.acquire() as conn:
            txn = Transaction(conn)
            async with txn:
                yield txn

    @asynccontextmanager
    async def acquire(self) -> AsyncIterator[asyncpg.Connection]:
        """Acquire a connection from the pool.

        Example:
            async with db.acquire() as conn:
                await conn.execute("...")
        """
        async with self.pool.acquire() as conn:
            yield conn



"""Derp ORM - A strongly-typed async Python ORM for PostgreSQL.

Example usage::

    from derp import Derp, Table, Field, eq
    from derp.fields import Serial, Varchar, Timestamp

    class User(Table, table="users"):
        id: int = Field(Serial(), primary_key=True)
        name: str = Field(Varchar(255))
        email: str = Field(Varchar(255), unique=True)
        created_at: datetime = Field(Timestamp(), default="now()")

    async with Derp("postgresql://...") as db:
        users = await db.select(User).where(User.c.name == "Alice").execute()
"""

from derp.config import DatabaseConfig
from derp.orm.engine import DatabaseEngine
from derp.orm.fields import (
    JSON,
    JSONB,
    UUID,
    Array,
    BigInt,
    BigSerial,
    Boolean,
    Char,
    Date,
    DoublePrecision,
    Field,
    FieldInfo,
    ForeignKey,
    ForeignKeyAction,
    Integer,
    Interval,
    Numeric,
    Real,
    Serial,
    SmallInt,
    Text,
    Time,
    Timestamp,
    Varchar,
)
from derp.orm.query import (
    ComparisonOperator,
    JoinType,
    LockMode,
    LogicalOperator,
    SortOrder,
    sql,
)
from derp.orm.table import Table

__version__ = "0.1.0"

__all__ = [
    # Main engine
    "DatabaseConfig",
    "DatabaseEngine",
    # Table definition
    "Table",
    "Field",
    "FieldInfo",
    "ForeignKey",
    "ForeignKeyAction",
    # Field types
    "Serial",
    "BigSerial",
    "SmallInt",
    "Integer",
    "BigInt",
    "Varchar",
    "Char",
    "Text",
    "Boolean",
    "Timestamp",
    "Date",
    "Time",
    "Interval",
    "Numeric",
    "Real",
    "DoublePrecision",
    "UUID",
    "JSON",
    "JSONB",
    "Array",
    # Query enums
    "JoinType",
    "LockMode",
    "SortOrder",
    "LogicalOperator",
    "ComparisonOperator",
    # SQL expressions
    "sql",
]



"""Queue client and typed models."""

from derp.config import CeleryConfig, QueueConfig, ScheduleConfig, VercelQueueConfig
from derp.queue.base import QueueClient, Schedule, ScheduleType, TaskState, TaskStatus
from derp.queue.celery import CeleryQueueClient
from derp.queue.exceptions import (
    QueueError,
    QueueNotConnectedError,
    QueueProviderError,
)
from derp.queue.vercel import VercelQueueClient

__all__ = [
    "CeleryConfig",
    "CeleryQueueClient",
    "QueueClient",
    "QueueConfig",
    "QueueError",
    "QueueNotConnectedError",
    "QueueProviderError",
    "Schedule",
    "ScheduleConfig",
    "ScheduleType",
    "TaskState",
    "TaskStatus",
    "VercelQueueClient",
    "VercelQueueConfig",
]
